package com.nordicpeak.flowengine.queries.basetablequery;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.interfaces.ImmutableAlternative;
import com.nordicpeak.flowengine.interfaces.QueryRequestProcessor;

public class TableQueryRequestProcessor<DropAlternative extends BaseDropDownAlternative, Column extends BaseTableColumn<DropAlternative, Column, Query>, Query extends BaseTableQuery<DropAlternative, Query, Column>, Row extends TableRow<DropAlternative, Column, Query>, CellValue extends TableCellValue<DropAlternative, Query, Row, Column>, QueryInstance extends BaseTableQueryInstance<DropAlternative, Row, Column, Query, CellValue>, QueryProvider extends TableQueryProviderModule<DropAlternative, Column, Query, Row, CellValue, QueryInstance, QueryProvider>> implements QueryRequestProcessor {

	private QueryInstance queryInstance;

	private final int maxHits;
	private final String showingPartialList1, showingPartialList2, showingPartialList3;

	public TableQueryRequestProcessor(QueryInstance queryInstance, int maxHits, String showingPartialList1, String showingPartialList2, String showingPartialList3) {

		this.queryInstance = queryInstance;
		this.maxHits = maxHits;
		this.showingPartialList1 = showingPartialList1;
		this.showingPartialList2 = showingPartialList2;
		this.showingPartialList3 = showingPartialList3;
	}

	@Override
	public void processRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String searchString = getSearchStringParameter(req);
		Integer columnID = NumberUtils.toInt(req.getParameter("columnID"));

		if (columnID == null) {

			sendEmptyJSONResponse(res);
			return;
		}

		Column column = queryInstance.getQuery().getColumnByID(columnID);

		if (column == null || column.getAlternativeList() == null) {

			sendEmptyJSONResponse(res);
			return;
		}

		List<? extends ImmutableAlternative> alternatives = column.getAlternativeList().getAlternatives();

		if (CollectionUtils.isEmpty(alternatives)) {

			sendEmptyJSONResponse(res);
			return;
		}

		JsonArray hitList = new JsonArray(maxHits + 1);

		if (!StringUtils.isEmpty(searchString)) {

			alternatives = CollectionUtils.filter(alternatives, alternative -> alternative.getName().toLowerCase().contains(searchString));
			appendAlternativesJSON(searchString, alternatives, hitList, column.isShowExactMatchesFirst());

			appendMoreHitsAlternative(alternatives, hitList);

			sendJSONResponse(hitList, res);
			
		} else if (column.isShowAlternativesOnEmptySearch()) {

			addAlternatives(alternatives, hitList);
			sendJSONResponse(hitList, res);

		}

	}

	private String getSearchStringParameter(HttpServletRequest req) throws UnsupportedEncodingException {

		String searchString = req.getParameter("q");

		if (StringUtils.isEmpty(searchString)) {

			return null;
		}

		searchString = URLDecoder.decode(searchString, "UTF-8");

		return searchString.toLowerCase();
	}

	private void appendAlternativesJSON(String searchString, List<? extends ImmutableAlternative> alternatives, JsonArray hitList, boolean showExactMatchesFirst) {

		List<Integer> addedSearchValues = new ArrayList<>();

		if (showExactMatchesFirst) {

			addExactMatchingSearchMatches(searchString, alternatives, hitList, addedSearchValues);
		}

		addContainingSearchMatches(searchString, alternatives, hitList, addedSearchValues);
	}

	private void addAlternatives(List<? extends ImmutableAlternative> alternatives, JsonArray hitList) {

		for (ImmutableAlternative alternative : alternatives) {

			JsonObject jsonAlternative = new JsonObject(2);

			addHits(hitList, alternative, jsonAlternative);

			if (hitList.size() == maxHits) {

				break;
			}
		}
	}

	private void addExactMatchingSearchMatches(String searchString, List<? extends ImmutableAlternative> alternatives, JsonArray hitList, List<Integer> addedAlternatives) {

		for (ImmutableAlternative alternative : alternatives) {

			JsonObject jsonAlternative = new JsonObject(2);

			if (alternative.getName().toLowerCase().equals(searchString)) {

				addHits(hitList, alternative, jsonAlternative);

				addedAlternatives.add(alternative.getAlternativeID());

				if (hitList.size() == maxHits) {

					break;
				}
			}
		}
	}

	private void addContainingSearchMatches(String searchString, List<? extends ImmutableAlternative> alternatives, JsonArray hitList, List<Integer> alreadyAddedAlternatives) {

		for (ImmutableAlternative alternative : alternatives) {

			JsonObject jsonAlternative = new JsonObject(2);

			if (alternative.getName().toLowerCase().contains(searchString) && !alreadyAddedAlternatives.contains(alternative.getAlternativeID())) {

				addHits(hitList, alternative, jsonAlternative);

				if (hitList.size() == maxHits) {

					break;
				}
			}
		}
	}

	private void addHits(JsonArray hitList, ImmutableAlternative alternative, JsonObject jsonAlternative) {

		jsonAlternative.putField("ID", alternative.getAlternativeID());
		jsonAlternative.putField("Name", alternative.getName());
		hitList.addNode(jsonAlternative);
	}

	private void appendMoreHitsAlternative(List<? extends ImmutableAlternative> alternatives, JsonArray hitList) {

		if (alternatives.size() > maxHits) {

			String moreHitsAlternativeName = showingPartialList1 + " " + hitList.size() + " " + showingPartialList2 + " " + alternatives.size() + " " + showingPartialList3;

			appendAlternativeJSON(hitList, "null", moreHitsAlternativeName);
		}
	}

	private void appendAlternativeJSON(JsonArray jsonArray, Object alternativeID, String alternativeName) {

		JsonObject jsonAlternative = new JsonObject(2);

		jsonAlternative.putField("ID", alternativeID);
		jsonAlternative.putField("Name", alternativeName);

		jsonArray.addNode(jsonAlternative);
	}

	protected static void sendEmptyJSONResponse(HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(1);
		jsonObject.putField("hitCount", "0");
		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	protected void sendJSONResponse(JsonArray jsonArray, HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(2);
		jsonObject.putField("hitCount", Integer.toString(jsonArray.size()));
		jsonObject.putField("hits", jsonArray);
		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	@Override
	public void close() throws Exception {}

}
