package com.nordicpeak.flowengine.biznodevehiclealternativelist;

import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.sql.DataSource;

import org.apache.log4j.Level;

import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.oauth.OAuthTokenException;
import se.unlogic.oauth.OAuthTokenHandler;
import se.unlogic.standardutils.dao.AnnotatedDAOFactory;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.populators.SwedishSocialSecurityPopulator;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.alternativelist.beans.AlternativeListDescriptor;
import com.nordicpeak.flowengine.alternativelist.interfaces.AlternativeListInstance;
import com.nordicpeak.flowengine.alternativelist.provider.BaseAlternativeListProvider;
import com.nordicpeak.flowengine.biznodevehiclealternativelist.biznoderesponse.BisnodeVehicleResponseObject;
import com.nordicpeak.flowengine.biznodevehiclealternativelist.biznoderesponse.VehicleHoldings;
import com.nordicpeak.flowengine.biznodevehiclealternativelist.biznoderesponse.VehicleType;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;

import flexjson.JSONDeserializer;

public class BiznodeVehicleAlternativeListProviderModule extends BaseAlternativeListProvider<ListSource, BiznodeVehicleAlternativeListInstance, ListSourceCRUD, BiznodeVehicleAlternativeListProviderModule, ListSourceEvent> {

	private static final SwedishSocialSecurityPopulator CITIZEN_ID_VALIDATOR = new SwedishSocialSecurityPopulator(true, false, true, true);

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "username", description = "")
	private String username;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "password", description = "")
	private String password;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection timeout", description = "The maximum time in seconds to wait for a connection", required = true)
	private Integer connectionTimeout = 10;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Request timeout", description = "The maximum time in seconds to wait for a response", required = true)
	private Integer requestTimeout = 10;
	
	private static final Integer MILLISECOND = 1000;

	private OAuthTokenHandler tokenHandler;

	private static final String BISNODE_TOKEN_OAUTH2 = "https://login.bisnode.com/as/token.oauth2";
	private static final String BISNODE_VEHICLE_API = "https://api.bisnode.com/vehicle-online/v1/person/";
	private static final String LIMITER = "/holdings?limit=300";
	private static final String GRANT_TYPE = "grant_type=client_credentials&scope=ts_online";

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, BiznodeVehicleAlternativeListProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		AnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);

		listDAO = daoFactory.getDAO(ListSource.class).getWrapper(Integer.class);
		listEventDAO = daoFactory.getDAO(ListSourceEvent.class);
		listEventListParamFactory = listEventDAO.getParamFactory(ListSourceEvent.LIST_RELATION, ListSource.class);

		listCRUD = new ListSourceCRUD(listDAO, new AnnotatedRequestPopulator<>(ListSource.class), this);
	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();
		if (ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.ERROR)) {

			tokenHandler = new OAuthTokenHandler(BISNODE_TOKEN_OAUTH2, GRANT_TYPE, connectionTimeout*MILLISECOND, requestTimeout*MILLISECOND, username, password, false);
		}
	}

	@Override
	public String getProviderName() {

		return moduleDescriptor.getDescription();
	}

	@Override
	public AlternativeListInstance getAlternativeListInstance(String name, AttributeHandler attributeHandler, User poster) {

		BiznodeVehicleAlternativeListInstance listInstance = getCachedInstance(name);

		if (listInstance != null) {

			return getAlternativeInstance(listInstance.getListSource(), attributeHandler, poster);
		}

		return null;
	}

	@Override
	public BiznodeVehicleAlternativeListInstance getAlternativeInstance(ListSource source, AttributeHandler attributeHandler, User poster) {

		BiznodeVehicleAlternativeListInstance listInstance = alternativeListCacheMap.get(source.getSourceID());

		if (listInstance != null) {

			List<ExternalAlternative> alternatives = getAlternatives(source, poster);

			if (alternatives == null) {
				return null;
			}

			listInstance = new BiznodeVehicleAlternativeListInstance(new AlternativeListDescriptor(providerID, source.getListName()), getAlternatives(source, poster), System.currentTimeMillis(), source);
		}

		return listInstance;
	}

	protected List<ExternalAlternative> getAlternatives(ListSource source, User poster) {

		log.info("Getting alternatives for list source " + source);

		if (username == null || password == null) {

			log.error("Empty username or password");
			return null;
		}

		try {
			String citizenID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(poster);

			if (citizenID == null) {
				return null;
			}

			citizenID = citizenID.replaceAll("[+-]", "");

			if (citizenID.length() == 12) {

				citizenID = citizenID.substring(2);
			}

			if (!CITIZEN_ID_VALIDATOR.validateDefaultFormat(citizenID)) {

				log.warn("Invalid citizen ID: " + citizenID);
				return null;
			}

			SimpleRequest holdingsRequest = setupRequest(citizenID);

			JSONDeserializer<BisnodeVehicleResponseObject> deserializer = new JSONDeserializer<>();

			// call Bisnode to get vehicles and deserialize response to java-object
			HTTPResponse response = HTTPUtils.sendHTTPGetRequest(holdingsRequest, StandardCharsets.UTF_8);
			BisnodeVehicleResponseObject restApiResponse = deserializer.deserialize(response.getValue(), BisnodeVehicleResponseObject.class);

			List<ExternalAlternative> alternatives = new ArrayList<>();

			List<VehicleHoldings> vehicleListins = restApiResponse.getData().getVehicleHoldings();

			if (vehicleListins != null) {
				int index = 1;
				for (VehicleHoldings vehicle : vehicleListins) {
					VehicleType type = VehicleType.getType(vehicle.getVehicleTypeCode());

					if (type != null && source.allowsVehicle(type)) {
						StringBuilder builder = new StringBuilder();
						builder.append(vehicle.getLicensePlate());
						builder.append("(");
						builder.append(vehicle.getMakeType());
						builder.append(")");
						ExternalAlternative alternative = new ExternalAlternative(index++, builder.toString());
						alternatives.add(alternative);
					}
				}
			}

			Collections.sort(alternatives);

			return alternatives;

		} catch (Exception e) {

			log.warn("Error getting alternatives for " + source, e);
		}

		return null;
	}

	private SimpleRequest setupRequest(String citizenIdentifier) throws OAuthTokenException {

		StringBuilder builder = new StringBuilder();
		builder.append(BISNODE_VEHICLE_API);
		builder.append(citizenIdentifier);
		builder.append(LIMITER);
		SimpleRequest holdingsRequest = new SimpleRequest(builder.toString());

		holdingsRequest.addHeader("Authorization", "Bearer " + tokenHandler.getAccessToken());

		holdingsRequest.setConnectionTimeout(connectionTimeout*MILLISECOND);
		holdingsRequest.setReadTimeout(requestTimeout*MILLISECOND);

		return holdingsRequest;
	}

	@Override
	protected BiznodeVehicleAlternativeListInstance getNewInstance(AlternativeListDescriptor descriptor, ListSource list) {

		return new BiznodeVehicleAlternativeListInstance(descriptor, null, System.currentTimeMillis(), list);
	}

	@Override
	protected ListSourceEvent getNewListEvent(ListSource list, Timestamp added, User poster, String message) {

		return new ListSourceEvent(list, added, poster, message);
	}

}
