package com.nordicpeak.flowengine.booking.cruds;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.utils.crud.FragmentLinkScriptFilter;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.object.ObjectUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;

import com.nordicpeak.flowengine.booking.BookingUtils;
import com.nordicpeak.flowengine.booking.CalendarAdminModule;
import com.nordicpeak.flowengine.booking.beans.BookingCalendar;
import com.nordicpeak.flowengine.booking.interfaces.CalendarFormExtensionProvider;

public class CalendarCRUD extends ModularCRUD<BookingCalendar, Integer, User, CalendarAdminModule> {

	protected final AnnotatedDAOWrapper<BookingCalendar, Integer> calendarDAOWrapper;

	public CalendarCRUD(AnnotatedDAOWrapper<BookingCalendar, Integer> bookingDAOWrapper, CalendarAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), bookingDAOWrapper, new AnnotatedRequestPopulator<>(BookingCalendar.class), "Calendar", "calendar", "", callback);

		this.calendarDAOWrapper = bookingDAOWrapper;

		setRequirePostForDelete(true);
	}

	@Override
	protected void appendListFormData(Document doc, Element listTypeElement, User user, HttpServletRequest req, URIParser uriParser, List<ValidationError> validationError) throws SQLException {

		super.appendListFormData(doc, listTypeElement, user, req, uriParser, validationError);

		if (callback.isEnableProfileSupport() && callback.getSiteProfileHandler() != null) {

			XMLUtils.append(doc, listTypeElement, "Profiles", callback.getSiteProfileHandler().getProfiles());
		}
	}

	@Override
	protected void appendAddFormData(Document doc, Element addTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendAddFormData(doc, addTypeElement, user, req, uriParser);

		appendFormData(doc, addTypeElement);

		appendExtensionViewFragments(null, doc, addTypeElement, user, req, uriParser);

		appendBean(new BookingCalendar(), addTypeElement, doc, user);
	}

	@Override
	protected void appendUpdateFormData(BookingCalendar calendar, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendUpdateFormData(calendar, doc, updateTypeElement, user, req, uriParser);

		appendManagerGroupsAndUsers(doc, updateTypeElement, calendar);

		appendExtraGroups(doc, updateTypeElement, calendar);

		appendFormData(doc, updateTypeElement);

		appendExtensionViewFragments(calendar, doc, updateTypeElement, user, req, uriParser);
	}

	protected void appendFormData(Document doc, Element element) {

		if (callback.isEnableProfileSupport() && callback.getSiteProfileHandler() != null) {

			XMLUtils.append(doc, element, "Profiles", callback.getSiteProfileHandler().getProfiles());
		}
	}

	protected void appendExtensionViewFragments(BookingCalendar calendar, Document doc, Element element, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		if (!CollectionUtils.isEmpty(callback.getExtensionProviders())) {

			Element pluginFragments = XMLUtils.appendNewElement(doc, element, "ExtensionFragments");

			for (CalendarFormExtensionProvider extensionProvider : callback.getExtensionProviders()) {

				ValidationException e = (ValidationException) req.getAttribute(extensionProvider.getProviderID());

				ViewFragment extensionSettings = extensionProvider.getCalendarFormViewFragment(calendar, req, user, uriParser, e);

				if (extensionSettings != null) {

					pluginFragments.appendChild(extensionSettings.toXML(doc));
					FragmentLinkScriptFilter.addViewFragment(extensionSettings, req);
				}
			}
		}
	}

	@Override
	protected void appendBean(BookingCalendar bean, Element targetElement, Document doc, User user) {

		callback.setCalendarTemplatedFields(bean);

		super.appendBean(bean, targetElement, doc, user);
	}

	@Override
	protected BookingCalendar populateFromAddRequest(HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		BookingCalendar calendar = null;

		try {
			calendar = super.populateFromAddRequest(req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		populate(calendar, req, user, uriParser, validationErrors);

		callback.filterHTML(calendar);

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return calendar;
	}

	@Override
	protected BookingCalendar populateFromUpdateRequest(BookingCalendar oldCalendar, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		ArrayList<ValidationError> validationErrors = new ArrayList<>();

		BookingCalendar calendar = oldCalendar;

		try {
			calendar = super.populateFromUpdateRequest(oldCalendar, req, user, uriParser);

		} catch (ValidationException e) {

			validationErrors.addAll(e.getErrors());
		}

		populate(calendar, req, user, uriParser, validationErrors);

		callback.filterHTML(calendar);

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		return calendar;
	}

	protected void populate(BookingCalendar calendar, HttpServletRequest req, User user, URIParser uriParser, List<ValidationError> validationErrors) {

		if (calendar != null) {

			callback.clearUnchangedTemplatedFields(calendar);

			clearFlowNotificationHandlerTemplatedFields(calendar);

			if (callback.isEnableProfileSupport() && calendar.getProfileID() == null) {

				validationErrors.add(new ValidationError("NoProfileChosen"));
			}

			if (calendar.isAllowDuplicateBookings() && calendar.isAllowPreliminaryBookings()) {

				validationErrors.add(new ValidationError("DuplicateAndPreliminaryMutualExclusivityViolation"));
			}

			if (!calendar.isSendUserRebookedManagerEmail() && !calendar.isSendRebookedGlobalEmail()) {

				calendar.setSendUserRebookedManagerEmail(true);
			}

			if (calendar.isBookingSizeSelectorEnabled() && calendar.getPrice() != null && calendar.getPrice() > 0) {

				calendar.setRestrictRebookingToCurrentBookingSize(true);
			}
		}

		if (extensionRequestsHasValidationErrors(req, user, uriParser)) {

			validationErrors.add(new ValidationError("PluginError"));
		}
	}

	private void clearFlowNotificationHandlerTemplatedFields(BookingCalendar calendar) {

		if (ObjectUtils.compare(calendar.getEmailSenderName(), callback.getEmailSenderName())) {

			calendar.setEmailSenderName(null);
		}

		if (ObjectUtils.compare(calendar.getEmailSenderAddress(), callback.getEmailSenderAddress())) {

			calendar.setEmailSenderAddress(null);
		}
	}

	private boolean extensionRequestsHasValidationErrors(HttpServletRequest req, User user, URIParser uriParser) {

		boolean hasErrors = false;

		if (!callback.getExtensionProviders().isEmpty()) {

			for (CalendarFormExtensionProvider extensionProvider : callback.getExtensionProviders()) {

				try {
					extensionProvider.validateCalendarFormRequest(req, user, uriParser);

				} catch (ValidationException e) {

					req.setAttribute(extensionProvider.getProviderID(), e);

					hasErrors = true;
				}
			}
		}

		return hasErrors;
	}

	@Override
	protected List<BookingCalendar> getAllBeans(User user, HttpServletRequest req, URIParser uriParser) throws SQLException {

		List<BookingCalendar> calendars = super.getAllBeans(user, req, uriParser);

		BookingUtils.setCalendarsTransientData(calendars, calendarDAOWrapper.getAnnotatedDAO());

		return calendars;
	}

	private void appendManagerGroupsAndUsers(Document doc, Element targetElement, BookingCalendar calendar) {

		Collection<Integer> managerUserIDs = calendar.getManagerUserIDs();

		if (!CollectionUtils.isEmpty(managerUserIDs)) {

			XMLUtils.append(doc, targetElement, "users", callback.getUserHandler().getUsers(managerUserIDs, false, true));
		}

		Collection<Integer> managerGroupIDs = calendar.getManagerGroupIDs();

		if (!CollectionUtils.isEmpty(managerGroupIDs)) {

			XMLUtils.append(doc, targetElement, "groups", callback.getGroupHandler().getGroups(managerGroupIDs, false));
		}
	}

	private void appendExtraGroups(Document doc, Element targetElement, BookingCalendar calendar) {

		Collection<Integer> adminGroupIDs = calendar.getAdminGroupIDs();

		if (!CollectionUtils.isEmpty(adminGroupIDs)) {

			XMLUtils.append(doc, targetElement, "adminGroups", callback.getGroupHandler().getGroups(adminGroupIDs, false));
		}
	}

	@Override
	protected ForegroundModuleResponse beanAdded(BookingCalendar calendar, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		processExtensionRequests(calendar, req, user, uriParser);

		return super.beanAdded(calendar, req, res, user, uriParser);
	}

	@Override
	protected ForegroundModuleResponse beanUpdated(BookingCalendar calendar, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		processExtensionRequests(calendar, req, user, uriParser);

		return super.beanUpdated(calendar, req, res, user, uriParser);
	}

	private void processExtensionRequests(BookingCalendar calendar, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		if (!callback.getExtensionProviders().isEmpty()) {

			for (CalendarFormExtensionProvider extensionProvider : callback.getExtensionProviders()) {

				extensionProvider.processCalendarFormRequest(calendar, req, user, uriParser);
			}
		}
	}

	@Override
	protected ForegroundModuleResponse beanDeleted(BookingCalendar bean, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		callback.getEventHandler().sendEvent(BookingCalendar.class, new CRUDEvent<>(CRUDAction.DELETE, bean), EventTarget.ALL);

		return super.beanDeleted(bean, req, res, user, uriParser);
	}
}
