package com.nordicpeak.flowengine.booking.cruds;

import java.lang.reflect.Field;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.utils.crud.FragmentLinkScriptFilter;
import se.unlogic.hierarchy.core.utils.crud.IntegerBeanIDParser;
import se.unlogic.hierarchy.core.utils.crud.ModularCRUD;
import se.unlogic.standardutils.annotations.WebPopulate;
import se.unlogic.standardutils.arrays.ArrayUtils;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AdvancedAnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.booking.BookingSlotAdminModule;
import com.nordicpeak.flowengine.booking.beans.BookingCalendar;
import com.nordicpeak.flowengine.booking.beans.BookingSlot;
import com.nordicpeak.flowengine.booking.beans.Repeating;
import com.nordicpeak.flowengine.booking.interfaces.BookingSlotFormExtensionProvider;

public class RepeatingCRUD extends ModularCRUD<Repeating, Integer, User, BookingSlotAdminModule> {

	protected final AdvancedAnnotatedDAOWrapper<Repeating, Integer> repeatingDAOWrapper;
	protected final AdvancedAnnotatedDAOWrapper<BookingSlot, Integer> bookingSlotDAOWrapper;

	public RepeatingCRUD(AdvancedAnnotatedDAOWrapper<Repeating, Integer> crudDAO, AdvancedAnnotatedDAOWrapper<BookingSlot, Integer> bookingSlotCrudDAO, BookingSlotAdminModule callback) {

		super(IntegerBeanIDParser.getInstance(), crudDAO, new AnnotatedRequestPopulator<>(Repeating.class), "Repeating", "repeating", "", callback);

		this.repeatingDAOWrapper = crudDAO;
		this.bookingSlotDAOWrapper = bookingSlotCrudDAO;

		setRequirePostForDelete(true);
	}

	@Override
	protected void checkUpdateAccess(Repeating repeating, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!hasAccess(repeating, user)) {

			throw new AccessDeniedException("User does not have update access for " + repeating);
		}
	}

	@Override
	protected void checkDeleteAccess(Repeating repeating, User user, HttpServletRequest req, URIParser uriParser) throws AccessDeniedException, URINotFoundException, SQLException {

		if (!hasAccess(repeating, user)) {

			throw new AccessDeniedException("User does not have delete access for " + repeating);
		}
	}

	public boolean hasAccess(Repeating repeating, User user) {

		return callback.hasCalendarAccess(repeating.getCalendar(), user);
	}

	@Override
	public ForegroundModuleResponse showUpdateForm(Repeating repeating, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, ValidationException validationException) throws Exception {

		repeating.setBookingSlots(callback.getBookingSlots(repeating, BookingSlot.MANAGERS_RELATION));

		return super.showUpdateForm(repeating, req, res, user, uriParser, validationException);
	}

	@Override
	protected void appendUpdateFormData(Repeating bean, Document doc, Element updateTypeElement, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		super.appendUpdateFormData(bean, doc, updateTypeElement, user, req, uriParser);

		appendExtensionViewFragments(bean, doc, updateTypeElement, user, req, uriParser);
	}

	protected void appendExtensionViewFragments(Repeating repeating, Document doc, Element element, User user, HttpServletRequest req, URIParser uriParser) throws Exception {

		if (!callback.getBookingSlotExtensionProviders().isEmpty()) {

			Element pluginFragments = XMLUtils.appendNewElement(doc, element, "ExtensionFragments");

			for (BookingSlotFormExtensionProvider extensionProvider : callback.getBookingSlotExtensionProviders()) {

				ValidationException validationException = (ValidationException) req.getAttribute(extensionProvider.getProviderID());

				ViewFragment extensionSettings = extensionProvider.getRepeatingFormViewFragment(repeating, req, user, uriParser, validationException);

				if (extensionSettings != null) {

					pluginFragments.appendChild(extensionSettings.toXML(doc));
					FragmentLinkScriptFilter.addViewFragment(extensionSettings, req);
				}
			}
		}
	}

	@Override
	protected Repeating populateFromUpdateRequest(Repeating repeating, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		Date oldToDate = repeating.getToDate();

		repeating = super.populateFromUpdateRequest(repeating, req, user, uriParser);

		if (repeating.getToDate().before(oldToDate)) {

			throw new ValidationException(new ValidationError("RepetitionToDateBeforeOldToDate"));
		}

		return repeating;
	}

	@Override
	protected void beforeUpdateFilteredBeanCommit(Repeating repeating, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		super.beforeUpdateFilteredBeanCommit(repeating, req, user, uriParser);

		log.info("User " + user + " updating booking slots for repeating " + repeating);

		LocalDateTime now = LocalDateTime.now();
		List<BookingSlot> bookingSlots = callback.getBookingSlots(repeating, BookingSlot.CALENDAR_RELATION, BookingSlot.REPEATING_RELATION, Repeating.CALENDAR_RELATION, BookingSlot.BOOKINGS_RELATION);

		if (!CollectionUtils.isEmpty(bookingSlots)) {

			List<ValidationError> validationErrors = new ArrayList<>();

			String description = validateParameter(BookingSlot.FIELD_DESCRIPTION, req, validationErrors);
			String location = validateParameter(BookingSlot.FIELD_LOCATION, req, validationErrors);
			List<User> managers = getManagers(req, repeating.getCalendar());

			ValidationException extensionValidationException = null;

			BookingSlot lastBookingSlot = null;
			for (BookingSlot bookingSlot : bookingSlots) {

				if (bookingSlot.getStart().toLocalDateTime().isAfter(now)) {

					bookingSlot.setLocation(location);
					bookingSlot.setDescription(description);
					bookingSlot.setManagers(managers);

					try {

						processExtensionRequests(bookingSlot, req, user, uriParser);

					} catch (ValidationException e) {

						extensionValidationException = e;
						continue;
					}

					bookingSlotDAOWrapper.update(bookingSlot, getTransactionHandler(req));
				}

				lastBookingSlot = bookingSlot;
			}

			if (extensionValidationException != null) {

				validationErrors.addAll(extensionValidationException.getErrors());
			}

			if (!validationErrors.isEmpty()) {

				throw new ValidationException(validationErrors);
			}

			createRepeatingBookingSlots(repeating, getTransactionHandler(req), lastBookingSlot, now, description, location, managers);
		}
	}

	private void createRepeatingBookingSlots(Repeating repeating, TransactionHandler transactionHandler, BookingSlot bookingSlot, LocalDateTime now, String description, String location, List<User> managers) throws SQLException {

		createRepeatingBookingSlots(repeating, bookingSlot, bookingSlotDAOWrapper, transactionHandler, now, description, location, managers);
	}

	public static void createRepeatingBookingSlots(Repeating repeating, TransactionHandler transactionHandler, BookingSlot bookingSlot, AdvancedAnnotatedDAOWrapper<BookingSlot, Integer> bookingSlotDAOWrapper, LocalDateTime now) throws SQLException {

		createRepeatingBookingSlots(repeating, bookingSlot, bookingSlotDAOWrapper, transactionHandler, now, null, null, null);
	}

	public static void createRepeatingBookingSlots(Repeating repeating, BookingSlot bookingSlot, AdvancedAnnotatedDAOWrapper<BookingSlot, Integer> bookingSlotDAOWrapper, TransactionHandler transactionHandler, LocalDateTime now, String overrideDescription, String overrideLocation, List<User> overrideManagers) throws SQLException {

		LocalDateTime startDate = bookingSlot.getStart().toLocalDateTime().plusDays(repeating.getRepetitionInterval());
		LocalDateTime endDate = bookingSlot.getEnd().toLocalDateTime().plusDays(repeating.getRepetitionInterval());

		while (!startDate.isAfter(repeating.getToDate().toLocalDate().atTime(LocalTime.MAX))) {

			if (startDate.isAfter(now)) {

				BookingSlot newBookingSlot = new BookingSlot(bookingSlot);
				newBookingSlot.setStart(Timestamp.valueOf(startDate));
				newBookingSlot.setEnd(Timestamp.valueOf(endDate));

				if (overrideLocation != null) {

					newBookingSlot.setLocation(overrideLocation);
				}

				if (overrideDescription != null) {

					newBookingSlot.setDescription(overrideDescription);
				}

				if (overrideManagers != null) {

					newBookingSlot.setManagers(overrideManagers);
				}

				bookingSlotDAOWrapper.add(newBookingSlot, transactionHandler);
			}

			startDate = startDate.plusDays(repeating.getRepetitionInterval());
			endDate = endDate.plusDays(repeating.getRepetitionInterval());
		}
	}

	private String validateParameter(Field parameterField, HttpServletRequest req, List<ValidationError> validationErrors) {

		WebPopulate webPopulate = parameterField.getAnnotation(WebPopulate.class);

		if (webPopulate == null) {

			return null;
		}

		String name = parameterField.getName();
		boolean required = webPopulate.required();
		Integer minLength = NumberUtils.toInt(webPopulate.minLength());
		Integer maxLength = NumberUtils.toInt(webPopulate.maxLength());

		return ValidationUtils.validateParameter(name, req, required, minLength, maxLength, validationErrors);
	}

	private void processExtensionRequests(BookingSlot bookingSlot, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		if (!callback.getBookingSlotExtensionProviders().isEmpty()) {

			boolean hasErrors = false;

			for (BookingSlotFormExtensionProvider extensionProvider : callback.getBookingSlotExtensionProviders()) {

				try {

					extensionProvider.processBookingSlotFormRequest(bookingSlot, req, user, uriParser);

				} catch (ValidationException e) {

					req.setAttribute(extensionProvider.getProviderID(), e);

					hasErrors = true;
				}
			}

			if (hasErrors) {

				throw new ValidationException(new ValidationError("ExtensionErrors"));
			}
		}
	}

	private List<User> getManagers(HttpServletRequest req, BookingCalendar calendar) {

		String[] managerIDs = req.getParameterValues("manager");

		List<User> managers = null;

		if (!ArrayUtils.isEmpty(managerIDs)) {

			Map<Integer, User> allowedManagers = callback.getAllowedManagers(calendar);

			if (!CollectionUtils.isEmpty(allowedManagers)) {

				managers = new ArrayList<>(managerIDs.length);

				for (String userIDString : managerIDs) {

					Integer userID = NumberUtils.toInt(userIDString);

					if (userID != null) {

						User manager = allowedManagers.get(userID);

						if (manager != null) {

							managers.add(manager);
						}
					}
				}
			}
		}

		return managers;
	}

	@Override
	protected void deleteFilteredBean(Repeating repeating, HttpServletRequest req, User user, URIParser uriParser) throws Exception {

		List<BookingSlot> bookingSlots = callback.getBookingSlots(repeating, BookingSlot.BOOKINGS_RELATION);

		int count = 0;
		boolean deletedAll = true;

		if (!CollectionUtils.isEmpty(bookingSlots)) {

			Date now = new Date();

			try (TransactionHandler transactionHandler = bookingSlotDAOWrapper.createTransaction()) {

				for (BookingSlot bookingSlot : bookingSlots) {

					if (CollectionUtils.isEmpty(bookingSlot.getBookings()) && bookingSlot.getStart().after(now)) {

						bookingSlotDAOWrapper.delete(bookingSlot, transactionHandler);

						callback.getEventHandler().sendEvent(BookingSlot.class, new CRUDEvent<>(CRUDAction.DELETE, bookingSlot), EventTarget.ALL);

						count++;

					} else {

						deletedAll = false;
					}
				}

				transactionHandler.commit();
			}
		}

		if (deletedAll) {

			log.info("Deleted all " + count + " booking slots for repeating " + repeating);

			super.deleteFilteredBean(repeating, req, user, uriParser);

		} else {

			log.info("Deleted " + count + " booking slots, not deleting repeating " + repeating + " due to remaining booking slots");
		}
	}

	@Override
	protected void redirectToListMethod(HttpServletRequest req, HttpServletResponse res, Repeating repeating) throws Exception {

		res.sendRedirect(req.getContextPath() + callback.getFullAlias() + "/list/" + repeating.getCalendar().getCalendarID());
	}

	@Override
	protected ForegroundModuleResponse filteredBeanAdded(Repeating repeating, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		processExtensionAddOrUpdate(repeating, req, user, uriParser, false);

		return super.filteredBeanAdded(repeating, req, res, user, uriParser);
	}

	@Override
	protected ForegroundModuleResponse filteredBeanUpdated(Repeating repeating, HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		processExtensionAddOrUpdate(repeating, req, user, uriParser, true);

		return super.filteredBeanUpdated(repeating, req, res, user, uriParser);
	}

	private void processExtensionAddOrUpdate(Repeating repeating, HttpServletRequest req, User user, URIParser uriParser, boolean processBookings) throws Exception {

		if (!callback.getBookingSlotExtensionProviders().isEmpty()) {

			List<BookingSlot> bookingSlots = callback.getBookingSlots(repeating, BookingSlot.CALENDAR_RELATION, BookingSlot.BOOKINGS_RELATION);

			if (!CollectionUtils.isEmpty(bookingSlots)) {

				for (BookingSlot bookingSlot : bookingSlots) {

					for (BookingSlotFormExtensionProvider extensionProvider : callback.getBookingSlotExtensionProviders()) {

						extensionProvider.processBookingSlotAddOrUpdate(bookingSlot, req, user, uriParser, processBookings);
					}
				}
			}
		}
	}

}
