package com.nordicpeak.flowengine.booking.schedule;

import java.sql.Timestamp;
import java.util.List;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.enums.SystemStatus;
import se.unlogic.hierarchy.core.interfaces.SystemInterface;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.dao.querys.ArrayListQuery;
import se.unlogic.standardutils.populators.IntegerPopulator;

import com.nordicpeak.flowengine.booking.beans.Booking;
import com.nordicpeak.flowengine.booking.beans.BookingEvent;
import com.nordicpeak.flowengine.booking.enums.BookingEventType;
import com.nordicpeak.flowengine.booking.events.BookingPassedEvent;

public class BookingPassedEventHandler implements Runnable {

	private HierarchyAnnotatedDAOFactory calledDaoFactory;
	private AnnotatedDAO<Booking> bookingDAO;
	private AnnotatedDAOWrapper<Booking, Integer> bookingDAOWrapper;

	private AnnotatedDAOWrapper<BookingEvent, Integer> eventDAOWrapper;

	private SystemInterface systemInterface;

	private Logger log = Logger.getLogger(BookingPassedEventHandler.class);

	public BookingPassedEventHandler(HierarchyAnnotatedDAOFactory calledDaoFactory, SystemInterface systemInterface) {

		this.calledDaoFactory = calledDaoFactory;
		this.systemInterface = systemInterface;
		bookingDAO = calledDaoFactory.getDAO(Booking.class);

		bookingDAOWrapper = bookingDAO.getWrapper(Integer.class);
		bookingDAOWrapper.addRelations(Booking.BOOKING_SLOT_RELATION);
		bookingDAOWrapper.setUseRelationsOnGet(true);

		eventDAOWrapper = calledDaoFactory.getDAO(BookingEvent.class).getWrapper(Integer.class);
	}

	@Override
	public void run() {

		try {

			// join tables bookings, booking_slots, booking_queries, booking_query_instances 
			// and get all bookingIDs for bookings that:
			// * is connected to a booking_query that is configured to send bookingsslotpassedevents
			// * has a enddate that has passed todays datetime plus added hours in bookings
			// * is connected to a booking that has not previously sent a bookingslotpassedevent

			//@formatter:off
			ArrayListQuery<Integer> idQuery = new ArrayListQuery<>(calledDaoFactory.getDataSource(),
					"select distinct b.bookingID from booking_slots as bs " +
					"inner join booking_query_instances as bqi on bs.bookingSlotID=bqi.bookingSlotID " +
					"inner join booking_queries as bq on bq.queryID=bqi.queryID " +
					"inner join bookings as b on b.bookingSlotID=bs.bookingSlotID " +
					"where bq.bookingSlotPassedStatusID is not null " + 
					"and NOW() >= DATE_ADD(bs.end, INTERVAL bq.bookingSlotPassedStatusTime HOUR) " + 
					"and b.bookingPassedEventSent=0;",
				IntegerPopulator.getPopulator());
			//@formatter:on

			List<Integer> IDs = idQuery.executeQuery();

			if (IDs == null) {

				return;
			}

			log.info("Found passed bookings");

			for (Integer id : IDs) {

				if (systemInterface.getSystemStatus() != SystemStatus.STARTED) {

					log.info("Aborting handling for passed bookings, waiting for system reboot...");
					return;
				}

				try (TransactionHandler transactionHandler = bookingDAO.createTransaction()) {

					log.info("Updating bookingpassedevent status on booking with bookingID " + id);

					Booking booking = bookingDAOWrapper.get(id);

					booking.setBookingPassedEventSent(true);
					bookingDAOWrapper.update(booking, transactionHandler);

					log.info("Sending event on Booking with bookingID " + id);
					BookingEvent bookingEvent = new BookingEvent(new Timestamp(System.currentTimeMillis()), null, BookingEventType.SLOT_PASSED, "", null);
					bookingEvent.setBookingSlot(booking);

					eventDAOWrapper.add(bookingEvent, transactionHandler);

					systemInterface.getEventHandler().sendEvent(Booking.class, new BookingPassedEvent(booking), EventTarget.ALL);
					transactionHandler.commit();
				}
			}

		} catch (Exception e) {

			log.error("Error sending bookingevents", e);
		}
	}
}
