package com.nordicpeak.flowengine.integration.castor;

import java.io.File;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.HTMLEditorSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.dao.RelationQuery;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.populators.StringPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.beans.Contact;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.FlowInstanceAttribute;
import com.nordicpeak.flowengine.beans.FlowInstanceEvent;
import com.nordicpeak.flowengine.events.StatusChangedByManagerEvent;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.integration.base.BaseIntegrationProviderModule;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationRequestException;
import com.nordicpeak.flowengine.integration.base.exceptions.IntegrationSettingsException;
import com.nordicpeak.flowengine.integration.castor.beans.AttachmentTypeMapping;
import com.nordicpeak.flowengine.integration.castor.beans.IntegrationEndpoint;
import com.nordicpeak.flowengine.integration.castor.beans.IntegrationSettings;
import com.nordicpeak.flowengine.integration.castor.beans.QueuedIntegrationRequest;
import com.nordicpeak.flowengine.integration.castor.client.CaseData;
import com.nordicpeak.flowengine.integration.castor.client.CastorClient;
import com.nordicpeak.flowengine.integration.castor.exceptions.IncompleteCaseDataException;
import com.nordicpeak.flowengine.integration.castor.exceptions.InvalidCaseIDException;
import com.nordicpeak.flowengine.integration.castor.exceptions.InvalidPartnerIDException;
import com.nordicpeak.flowengine.integration.castor.proxy.CaseResult;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.PDFProvider;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;
import com.nordicpeak.flowengine.notifications.StandardFlowNotificationHandler;
import com.nordicpeak.flowengine.utils.FlowInstanceUtils;
import com.nordicpeak.flowengine.utils.PDFByteAttachment;

public class CastorIntegrationProviderModule extends BaseIntegrationProviderModule<QueuedIntegrationRequest, IntegrationSettings, IntegrationEndpoint> {

	private static final String SUPPORTED_TAGS;

	static {
		ArrayList<String> tagList = new ArrayList<>(40);

		tagList.add(StandardFlowNotificationHandler.CONTACT_TAG_SOURCE_FACTORY.getAvailableTags());
		tagList.add(StandardFlowNotificationHandler.FLOW_TAG_SOURCE_FACTORY.getAvailableTags());
		tagList.add(StandardFlowNotificationHandler.FLOWINSTANCE_TAG_SOURCE_FACTORY.getAvailableTags());
		tagList.add(StandardFlowNotificationHandler.STATUS_TAG_SOURCE_FACTORY.getAvailableTags());

		SUPPORTED_TAGS = StringUtils.toCommaSeparatedString(tagList);
	}
	
	private static final AnnotatedRequestPopulator<IntegrationSettings> POPULATOR = new AnnotatedRequestPopulator<>(IntegrationSettings.class);
	private static final String STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME = "castorIntegrationStatusRequestEventID";
	private static final RelationQuery UPDATE_FLOW_INSTANCE_RELATIONS = new RelationQuery(FlowInstance.ATTRIBUTES_RELATION);

	@XSLVariable(name = "i18n.QueryName")
	protected String queryNameDisplayName;

	@XSLVariable(name = "i18n.AttachmentTypeID")
	protected String attachmentTypeIDDisplayName;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Error email subject", description = "The subject of emails sent when requests fail")
	@XSLVariable(prefix = "java.")
	private String errorEmailSubject = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Error email message", description = "The message of emails sent when errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String errorEmailMessage = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Invalid partner ID error email message", description = "The message of emails sent when invalid partner ID errors are received", required = true)
	@XSLVariable(prefix = "java.")
	private String invalidPartnerIDErrorEmailMessage = "Not set";

	@ModuleSetting
	@HTMLEditorSettingDescriptor(name = "Invalid case ID error email message", description = "The message of emails sent when create cases are returned without valid case ID", required = true)
	@XSLVariable(prefix = "java.")
	private String invalidCaseIDErrorEmailMessage = "Not set";

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Log generated template", description = "Toggles logging of the generated CastorCaseTemplate")
	protected boolean logTemplate;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Disable certificate validation", description = "Controls whether certificate should be validated when using SSL or not")
	private boolean disableCertificateValidation = false;

	@InstanceManagerDependency(required = true)
	private PDFProvider pdfProvider;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, this.getClass().getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		SimpleAnnotatedDAOFactory daoFactory = new SimpleAnnotatedDAOFactory(dataSource);

		settingsDAO = daoFactory.getDAO(IntegrationSettings.class);
		endpointDAO = daoFactory.getDAO(IntegrationEndpoint.class);
		queuedRequestDAO = daoFactory.getDAO(QueuedIntegrationRequest.class);

		flowFamilyIDParamFactory = settingsDAO.getParamFactory("flowFamilyID", Integer.class);
		endpointIDParamFactory = endpointDAO.getParamFactory("endpointID", Integer.class);
		lastRetryParamFactory = queuedRequestDAO.getParamFactory("lastRetry", Timestamp.class);
		eventIDParamFactory = queuedRequestDAO.getParamFactory("eventID", Integer.class);
		flowInstanceIDParamFactory = queuedRequestDAO.getParamFactory("flowInstanceID", Integer.class);
	}

	@EventListener(channel = FlowInstanceManager.class, priority = 49)
	public void processEvent(SubmitEvent event, EventSource eventSource) throws Exception {

		FlowInstanceManager flowInstanceManager = event.getFlowInstanceManager();
		ImmutableFlowInstance flowInstance = flowInstanceManager.getFlowInstance();
		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled() || !shouldProcessEvent(event)) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = event.getEvent();

		processEvent(flowInstanceEvent, settings);
	}

	private void processEvent(FlowInstanceEvent flowInstanceEvent, IntegrationSettings settings) throws Exception {

		FlowInstance flowInstance = flowInstanceEvent.getFlowInstance();

		try {

			processIntegrationEvent(flowInstance, flowInstanceEvent);

		} catch (IncompleteCaseDataException e) {

			if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (InvalidPartnerIDException e) {

			if (maxRetryCount > 0) {

				log.info("Adding request to retry queue");

				addQueuedEvent(flowInstanceEvent, flowInstance);
			}

			if (settings.isSendConfigurationErrorEmail()) {

				log.info("Sending invalid partner id error emails");

				sendEmails(settings.getConfigurationErrorEmailAddresses(), errorEmailSubject, invalidPartnerIDErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (InvalidCaseIDException e) {

			if (settings.isSendInvalidCaseIDErrorEmail()) {

				log.info("Sending invalid case id error emails");

				sendEmails(settings.getInvalidCaseIDErrorEmailAddresses(), errorEmailSubject, invalidCaseIDErrorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationRequestException e) {

			if (maxRetryCount > 0) {

				log.info("Adding request to retry queue");

				addQueuedEvent(flowInstanceEvent, flowInstance);

			} else if (settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}

		} catch (IntegrationSettingsException e) {

			log.error("Settings error while processing event " + flowInstanceEvent, e);
		}
	}

	@Override
	public void retryEvent(FlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler) throws Exception {

		try {

			super.retryEvent(flowInstance, event, queuedRequest, transactionHandler);

		} catch (IncompleteCaseDataException e) {

			log.info("Deleting " + queuedRequest + " due to previous error");

			queuedRequestDAO.delete(queuedRequest, transactionHandler);

			IntegrationSettings settings;

			if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null && settings.isSendErrorEmail()) {

				log.info("Sending error emails");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));
			}
		}
	}

	@Override
	public void processIntegrationEvent(FlowInstance flowInstance, FlowInstanceEvent event) throws Exception {

		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow());

		if (settings == null) {

			throw new IntegrationSettingsException("Settings are null");
		}

		if (!settings.isEnabled()) {

			throw new IntegrationSettingsException("Settings are disabled");
		}

		if (shouldDelayRequest(flowInstance, event, settings)) {

			return;
		}

		CastorClient client = new CastorClient(settings, connectionTimeout, requestTimeout, logTemplate, disableCertificateValidation);

		TagReplacer tagReplacer = new TagReplacer();

		tagReplacer.addTagSource(StandardFlowNotificationHandler.FLOWINSTANCE_TAG_SOURCE_FACTORY.getTagSource(flowInstance));
		tagReplacer.addTagSource(StandardFlowNotificationHandler.FLOW_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getFlow()));
		tagReplacer.addTagSource(StandardFlowNotificationHandler.STATUS_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getStatus()));

		Contact posterContact = FlowInstanceUtils.getPosterContact(flowInstance);

		if (posterContact != null) {
			tagReplacer.addTagSource(StandardFlowNotificationHandler.CONTACT_TAG_SOURCE_FACTORY.getTagSource(posterContact));
		}

		File pdfFile = pdfProvider.getPDF(flowInstance.getFlowInstanceID(), event.getEventID());

		List<PDFByteAttachment> pdfAttachments = null;

		if (settings.isAddAttachmentsSeparately()) {

			pdfAttachments = pdfProvider.getPDFAttachments(pdfFile, true);
		}

		CaseData caseData = new CaseData(settings, flowInstance, pdfFile, pdfAttachments, tagReplacer);

		try {

			CaseResult caseResult = client.createCastorCase(caseData);

			//Store external ID on flow instance
			if(flowInstance.getAttributeHandler() != null) {
				
				flowInstance.getAttributeHandler().setAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, caseResult.getCaseNumber());
			}
			
			//Store external ID in DB
			FlowInstanceAttribute attribute = new FlowInstanceAttribute(Constants.FLOW_INSTANCE_EXTERNAL_ID_ATTRIBUTE, caseResult.getCaseNumber());
			attribute.setFlowInstance(flowInstance);

			//TODO set store guid as hidden attribute

			flowAdminModule.getDAOFactory().getFlowInstanceAttributeDAO().add(attribute);
			
			log.info("Successfully created Castor case (caseID: " + caseResult.getCaseID() + ", caseNumber: " + caseResult.getCaseNumber() + ") for event " + event + " from flow instance " + flowInstance);

		} catch (IncompleteCaseDataException e) {

			boolean shouldLogAsError = !settings.isSendErrorEmail();

			logAndRethrow(flowInstance, event, settings, e, shouldLogAsError);

		} catch (InvalidPartnerIDException e) {

			boolean shouldLogAsError = maxRetryCount == 0 && !settings.isSendConfigurationErrorEmail();

			logAndRethrow(flowInstance, event, settings, e, shouldLogAsError);

		} catch (InvalidCaseIDException e) {

			boolean shouldLogAsError = !settings.isSendInvalidCaseIDErrorEmail();

			logAndRethrow(flowInstance, event, settings, e, shouldLogAsError);

		} catch (IntegrationRequestException e) {

			boolean shouldLogAsError = maxRetryCount == 0 && !settings.isSendErrorEmail();

			logAndRethrow(flowInstance, event, settings, e, shouldLogAsError);
		}
	}

	private boolean shouldDelayRequest(FlowInstance flowInstance, FlowInstanceEvent event, IntegrationSettings settings) throws SQLException {

		if (!settings.isCreateCaseOnSpecificStatus() || flowInstance.getStatus().getName().equalsIgnoreCase(settings.getCaseStatusName())) {

			return false;
		}

		log.info("Delaying integration request for event " + event + ", flow instance status " + flowInstance.getStatus() + " is not equal to set status " + settings.getCaseStatusName());

		flowInstance.getAttributeHandler().setAttribute(STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME, event.getEventID());

		updateFlowInstance(flowInstance);

		return true;
	}

	private void logAndRethrow(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, IntegrationSettings integrationSettings, IntegrationRequestException e, boolean shouldLogAsError) throws IntegrationRequestException {

		logResponse("Error creating Castor case for event " + event + " from flow instance " + flowInstance + " to endpoint " + integrationSettings.getEndpoint(), e, shouldLogAsError);

		throw e;
	}

	@EventListener(channel = FlowInstance.class)
	public void processEvent(StatusChangedByManagerEvent event, EventSource eventSource) throws Exception {

		FlowInstance flowInstance = event.getFlowInstance();
		MutableAttributeHandler attributeHandler = flowInstance.getAttributeHandler();
		Integer eventID = attributeHandler.getInt(STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME);

		if (eventID == null) {

			return;
		}

		FlowInstanceEvent flowInstanceEvent = flowAdminModule.getFlowInstanceEvent(eventID, getFlowInstanceRelations());

		if (flowInstanceEvent == null) {

			return;
		}

		IntegrationSettings settings = getIntegrationSettings(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (settings == null || !settings.isEnabled() || !settings.isCreateCaseOnSpecificStatus()) {

			return;
		}

		if (!flowInstance.getStatus().getName().equalsIgnoreCase(settings.getCaseStatusName())) {

			return;
		}

		log.info("Processing integration request for event " + flowInstanceEvent + " due to changed status (" + flowInstance.getStatus() + ")");

		processEvent(flowInstanceEvent, settings);

		attributeHandler.removeAttribute(STATUS_REQUEST_EVENT_ID_ATTRIBUTE_NAME);

		updateFlowInstance(flowInstance);
	}

	@WebPublic(alias = "update")
	public ForegroundModuleResponse updateFlowIntegrationSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);

		if (flow == null) {

			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));
		}
		
		if (!flowAdminModule.hasFlowAccess(user, flow)) {

			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}

		IntegrationSettings integrationSettings = getIntegrationSettings(flow);

		ValidationException validationException = null;

		if (req.getMethod().equalsIgnoreCase("POST")) {

			try {

				updateIntegrationSettings(integrationSettings, flow, user, req);

				flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + getShowFlowModuleAnchorLink());

			} catch (ValidationException e) {

				validationException = e;
			}
		}

		log.info("User " + user + " requesting update integration settings form for flow " + flow);

		Document doc = createDocument(req, uriParser, user);

		Element integrationSettingsElement = doc.createElement("UpdateIntegrationSettings");
		doc.getDocumentElement().appendChild(integrationSettingsElement);

		integrationSettingsElement.appendChild(flow.toXML(doc));

		XMLUtils.append(doc, integrationSettingsElement, integrationSettings);
		XMLUtils.append(doc, integrationSettingsElement, "Endpoints", endpointDAO.getAll());

		XMLUtils.appendNewElement(doc, integrationSettingsElement, "SupportedTags", SUPPORTED_TAGS);
		
		if (validationException != null) {

			integrationSettingsElement.appendChild(validationException.toXML(doc));
			integrationSettingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}

		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}

	@WebPublic(alias = "statuses")
	public ForegroundModuleResponse getStatuses(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		return flowAdminModule.getStatuses(req, res, user, uriParser);
	}

	private void updateIntegrationSettings(IntegrationSettings integrationSettings, Flow flow, User user, HttpServletRequest req) throws SQLException, ValidationException {

		if (integrationSettings == null) {

			integrationSettings = new IntegrationSettings();
			integrationSettings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
		}

		List<ValidationError> validationErrors = new ArrayList<>();

		POPULATOR.populate(integrationSettings, req);

		IntegrationEndpoint endpoint = getEndpoint(req);

		if (endpoint == null) {

			validationErrors.add(new ValidationError("NoEndpointSelected"));

		} else {

			integrationSettings.setEndpoint(endpoint);
		}

		if (integrationSettings.isAddAttachmentsSeparately()) {

			integrationSettings.setAttachmentTypeMappings(populateAttachmentTypeMappings(req, validationErrors));
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}

		log.info("User " + user + " updating integration settings for flow " + flow);

		if (integrationSettings.isPopulated()) {

			settingsDAO.addOrUpdate(integrationSettings, null);

		} else {

			settingsDAO.delete(integrationSettings);
		}
	}

	private List<AttachmentTypeMapping> populateAttachmentTypeMappings(HttpServletRequest req, List<ValidationError> validationErrors) {

		String[] attachmentTypeMappingIDs = req.getParameterValues("attachmentTypeMappingID");

		if (attachmentTypeMappingIDs != null) {

			List<AttachmentTypeMapping> mappings = new ArrayList<>(attachmentTypeMappingIDs.length);

			for (String attachmentTypeMappingID : attachmentTypeMappingIDs) {

				String queryName = ValidationUtils.validateParameter("queryName-" + attachmentTypeMappingID, queryNameDisplayName, req, true, 1, 255, StringPopulator.getPopulator(), validationErrors);
				Integer attachmentTypeID = ValidationUtils.validateParameter("attachmentTypeID-" + attachmentTypeMappingID, attachmentTypeIDDisplayName, req, true, 1, 255, IntegerPopulator.getPopulator(), validationErrors);

				AttachmentTypeMapping mapping = new AttachmentTypeMapping();
				mapping.setQueryName(queryName);
				mapping.setAttachmentTypeID(attachmentTypeID);

				mappings.add(mapping);
			}

			return mappings;
		}

		return null;
	}

	@Override
	protected QueuedIntegrationRequest getNewQueuedIntegrationRequest() {

		return new QueuedIntegrationRequest();
	}

	@Override
	protected void handleMaxRetryCountExceeded(ImmutableFlowInstance flowInstance, FlowInstanceEvent event, QueuedIntegrationRequest queuedRequest, TransactionHandler transactionHandler, IntegrationRequestException lastRequestException) throws SQLException {

		super.handleMaxRetryCountExceeded(flowInstance, event, queuedRequest, transactionHandler, lastRequestException);

		IntegrationSettings settings;

		if (flowInstance != null && (settings = getIntegrationSettings(flowInstance.getFlow())) != null) {

			if (settings.isSendErrorEmail()) {

				log.info("Sending error emails due to max retries exceeded");

				sendEmails(settings.getErrorEmailAddresses(), errorEmailSubject, errorEmailMessage, flowInstance, getTagReplacer(flowInstance));

			} else {

				log.error(queuedRequest + " hit its max retry count and error email is disabled");
			}
		}
	}

	@Override
	protected String getShowFlowModuleAnchorLink() {

		return "#castorintegrationsettings";
	}

	private void updateFlowInstance(FlowInstance flowInstance) throws SQLException {

		flowAdminModule.getDAOFactory().getFlowInstanceDAO().update(flowInstance, UPDATE_FLOW_INSTANCE_RELATIONS);
	}
}
