package com.nordicpeak.flowengine.deduplication;

import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.annotations.XSLVariable;
import se.unlogic.hierarchy.core.beans.LinkTag;
import se.unlogic.hierarchy.core.beans.ScriptTag;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.CRUDAction;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.enums.EventTarget;
import se.unlogic.hierarchy.core.events.CRUDEvent;
import se.unlogic.hierarchy.core.exceptions.AccessDeniedException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.ModuleViewFragmentTransformer;
import se.unlogic.hierarchy.core.utils.ViewFragmentModule;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.querys.ArrayListQuery;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.populators.IntegerPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.NonNegativeStringIntegerValidator;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationErrorType;
import se.unlogic.standardutils.validation.ValidationException;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;
import se.unlogic.webutils.populators.annotated.AnnotatedRequestPopulator;
import se.unlogic.webutils.validation.ValidationUtils;

import com.nordicpeak.flowengine.FlowAdminModule;
import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowAdminExtensionShowView;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.dao.FlowEngineDAOFactory;
import com.nordicpeak.flowengine.deduplication.beans.FlowFamilyDeduplicationSettings;
import com.nordicpeak.flowengine.enums.EventType;
import com.nordicpeak.flowengine.events.SubmitEvent;
import com.nordicpeak.flowengine.interfaces.FlowAdminExtensionViewProvider;
import com.nordicpeak.flowengine.managers.FlowInstanceManager;

public class DeduplicationModule extends AnnotatedForegroundModule implements ViewFragmentModule<ForegroundModuleDescriptor>, FlowAdminExtensionViewProvider {
	
	private static final AnnotatedRequestPopulator<FlowFamilyDeduplicationSettings> POPULATOR = new AnnotatedRequestPopulator<FlowFamilyDeduplicationSettings>(FlowFamilyDeduplicationSettings.class);
	
	@XSLVariable(prefix = "java.")
	private String extensionViewTitle = "Deduplication";
	
	@XSLVariable(prefix = "java.")
	private String flowEventDescription = "Deduplicated";
	
	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Priority", description = "The priority of this extension provider compared to other providers. A low value means a higher priority. Valid values are 0 - " + Integer.MAX_VALUE + ".", required = true, formatValidator = NonNegativeStringIntegerValidator.class)
	protected int extensionPriority = 0;
	
	protected FlowEngineDAOFactory daoFactory;
	
	protected AnnotatedDAOWrapper<FlowFamilyDeduplicationSettings, Integer> flowFamilyDeduplicationSettingsDAOWrapper;
	
	protected QueryParameterFactory<Status, Flow> statusFlowParamFactory;
	protected QueryParameterFactory<Status, String> statusNameParamFactory;
	
	protected ModuleViewFragmentTransformer<ForegroundModuleDescriptor> viewFragmentTransformer;
	
	protected FlowAdminModule flowAdminModule;
	
	@InstanceManagerDependency(required = true)
	private StaticContentModule staticContentModule;
	
	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {
		
		super.init(moduleDescriptor, sectionInterface, dataSource);
		
		viewFragmentTransformer = new ModuleViewFragmentTransformer<ForegroundModuleDescriptor>(sectionInterface.getForegroundModuleXSLTCache(), this, systemInterface.getEncoding());
	}
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {
		
		super.createDAOs(dataSource);
		
		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, DeduplicationModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));
		
		if (upgradeResult.isUpgrade()) {
			
			log.info(upgradeResult.toString());
		}
		
		daoFactory = new FlowEngineDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler());
		
		statusFlowParamFactory = daoFactory.getStatusDAO().getParamFactory("flow", Flow.class);
		statusNameParamFactory = daoFactory.getStatusDAO().getParamFactory("name", String.class);
		
		HierarchyAnnotatedDAOFactory simpleDAOFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface.getUserHandler(), systemInterface.getGroupHandler(), false, true, false);
		
		flowFamilyDeduplicationSettingsDAOWrapper = simpleDAOFactory.getDAO(FlowFamilyDeduplicationSettings.class).getWrapper(Integer.class);
	}
	
	@Override
	public void unload() throws Exception {
		
		super.unload();
	}
	
	@InstanceManagerDependency(required = true)
	public void setFlowAdminModule(FlowAdminModule flowAdminModule) {
		
		if (flowAdminModule == null && this.flowAdminModule != null) {
			
			this.flowAdminModule.removeExtensionViewProvider(this);
		}
		
		this.flowAdminModule = flowAdminModule;
		
		if (this.flowAdminModule != null) {
			
			this.flowAdminModule.addExtensionViewProvider(this);
		}
	}
	
	public Document createDocument(HttpServletRequest req, URIParser uriParser) {
		
		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");
		doc.appendChild(documentElement);
		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(this.moduleDescriptor.toXML(doc));
		
		XMLUtils.appendNewElement(doc, documentElement, "ModuleURI", req.getContextPath() + getFullAlias());
		XMLUtils.appendNewElement(doc, documentElement, "StaticContentURL", staticContentModule.getModuleContentURL(moduleDescriptor));
		
		return doc;
	}
	
	@Override
	public ForegroundModuleDescriptor getModuleDescriptor() {
		return moduleDescriptor;
	}
	
	@Override
	public List<LinkTag> getLinkTags() {
		
		return links;
	}
	
	@Override
	public List<ScriptTag> getScriptTags() {
		
		return scripts;
	}
	
	@EventListener(channel = FlowInstanceManager.class, priority = 60)
	public void processEvent(SubmitEvent event, EventSource source) {
		
		try {
			if (source.isLocal() && event.getEvent().getEventType() == EventType.SUBMITTED) {
				
				Timestamp now = event.getEvent().getAdded();
				User user = event.getEvent().getPoster();
				
				FlowInstanceManager flowInstanceManager = event.getFlowInstanceManager();
				
				FlowFamilyDeduplicationSettings settings = flowFamilyDeduplicationSettingsDAOWrapper.get(flowInstanceManager.getFlowInstance().getFlow().getFlowFamily().getFlowFamilyID());
				
				if (settings != null) {
					
					if (settings.isMatchPoster() && user == null) {
						return;
					}
					
					List<Integer> flowInstanceIDs = null;
					
					if (settings.getAttribute() != null) {
						
						String attributeValue = event.getFlowInstanceManager().getFlowInstance().getAttributeHandler().getString(settings.getAttribute());
						
						if (!StringUtils.isEmpty(attributeValue)) {
							
							if (settings.isMatchPoster()) {
								
								StringBuilder sql = new StringBuilder();
								sql.append("SELECT flowengine_flow_instances.flowInstanceID " +
										"FROM flowengine_flow_instances " +
										"INNER JOIN flowengine_flows ON flowengine_flows.flowID = flowengine_flow_instances.flowID " +
										"INNER JOIN flowengine_flow_instance_attributes ON flowengine_flow_instance_attributes.flowInstanceID = flowengine_flow_instances.flowInstanceID " +
										"WHERE " +
										"flowengine_flow_instances.firstSubmitted IS NOT NULL " +
										"AND flowengine_flow_instances.poster = ? " +
										"AND flowengine_flows.flowFamilyID = ? " +
										"AND flowengine_flow_instance_attributes.name = ? " +
										"AND flowengine_flow_instance_attributes.value = ?");
								
								ArrayListQuery<Integer> query = new ArrayListQuery<Integer>(dataSource, sql.toString(), IntegerPopulator.getPopulator());
							
								query.setInt(1, user.getUserID());
								query.setInt(2, event.getFlowInstanceManager().getFlowInstance().getFlow().getFlowFamily().getFlowFamilyID());
								query.setString(3, settings.getAttribute());
								query.setString(4, attributeValue);
								
								flowInstanceIDs = query.executeQuery();
								
							} else {
								
								StringBuilder sql = new StringBuilder();
								sql.append("SELECT flowengine_flow_instances.flowInstanceID " +
										"FROM flowengine_flow_instances " +
										"INNER JOIN flowengine_flows ON flowengine_flows.flowID = flowengine_flow_instances.flowID " +
										"INNER JOIN flowengine_flow_instance_attributes ON flowengine_flow_instance_attributes.flowInstanceID = flowengine_flow_instances.flowInstanceID " +
										"WHERE " +
										"flowengine_flow_instances.firstSubmitted IS NOT NULL " +
										"AND flowengine_flows.flowFamilyID = ? " +
										"AND flowengine_flow_instance_attributes.name = ? " +
										"AND flowengine_flow_instance_attributes.value = ?");
								
								ArrayListQuery<Integer> query = new ArrayListQuery<Integer>(dataSource, sql.toString(), IntegerPopulator.getPopulator());
								query.setInt(1, event.getFlowInstanceManager().getFlowInstance().getFlow().getFlowFamily().getFlowFamilyID());
								query.setString(2, settings.getAttribute());
								query.setString(3, attributeValue);
								
								flowInstanceIDs = query.executeQuery();
							}
						}
						
					} else if (settings.isMatchPoster()) {
						
						StringBuilder sql = new StringBuilder();
						sql.append("SELECT flowengine_flow_instances.flowInstanceID " +
								"FROM flowengine_flow_instances " +
								"INNER JOIN flowengine_flows ON flowengine_flows.flowID = flowengine_flow_instances.flowID " +
								"WHERE " +
								"flowengine_flow_instances.firstSubmitted IS NOT NULL " +
								"AND flowengine_flow_instances.poster = ? " +
								"AND flowengine_flows.flowFamilyID = ? ");
						
						ArrayListQuery<Integer> query = new ArrayListQuery<Integer>(dataSource, sql.toString(), IntegerPopulator.getPopulator());
						query.setInt(1, event.getEvent().getPoster().getUserID());
						query.setInt(2, event.getFlowInstanceManager().getFlowInstance().getFlow().getFlowFamily().getFlowFamilyID());
						
						flowInstanceIDs = query.executeQuery();
					}
					
					if (!CollectionUtils.isEmpty(flowInstanceIDs)) {
						
						flowInstanceIDs.remove(event.getFlowInstanceManager().getFlowInstanceID());
						
						if (settings.getEndDate() != null) {
							
							settings.setEndDate(DateUtils.setTimeToMaximum(settings.getEndDate()));
						}
						
						for (Integer flowInstanceID : flowInstanceIDs) {
							
							FlowInstance flowInstance = flowAdminModule.getFlowInstance(flowInstanceID);
							
							if (!flowInstance.getStatus().getName().equalsIgnoreCase(settings.getStatusName())) {
								
								if (settings.getStartDate() != null && flowInstance.getFirstSubmitted().before(settings.getStartDate())){
									continue;
								}
								
								if (settings.getEndDate() != null && flowInstance.getFirstSubmitted().after(settings.getEndDate())){
									continue;
								}
								
								Status status = getStatus(flowInstance.getFlow(), settings.getStatusName());
								
								if (status == null) {
									
									log.warn("Flow " + flowInstance.getFlow() + " does not have a status with name " + settings.getStatusName());
									continue;
								}
								
								log.info("Found duplicate submitted flow instance " + flowInstance + ", changing status to " + status);

								flowInstance.setStatus(status);
								flowInstance.setLastStatusChange(now);
								
								daoFactory.getFlowInstanceDAO().update(flowInstance);
								
								flowAdminModule.getFlowInstanceEventGenerator().addFlowInstanceEvent(flowInstance, EventType.STATUS_UPDATED, flowEventDescription, user);
								
								systemInterface.getEventHandler().sendEvent(FlowInstance.class, new CRUDEvent<FlowInstance>(CRUDAction.UPDATE, flowInstance), EventTarget.ALL);
							}
						}
					}
				}
			}
			
		} catch (Exception e) {
			
			log.error("Error ", e);
		}
	}
	
	private Status getStatus(Flow flow, String name) throws SQLException {
		
		HighLevelQuery<Status> query = new HighLevelQuery<Status>();
		
		query.addParameter(statusFlowParamFactory.getParameter(flow));
		query.addParameter(statusNameParamFactory.getParameter(name));
		
		return daoFactory.getStatusDAO().get(query);
	}
	
	
	@Override
	public int getPriority() {
		return extensionPriority;
	}
	
	@Override
	public FlowAdminExtensionShowView getShowView(Flow flow, HttpServletRequest req, User user, URIParser uriParser) throws TransformerConfigurationException, TransformerException, SQLException {
		
		Document doc = createDocument(req, uriParser);
		
		Element showViewElement = doc.createElement("ShowSettings");
		doc.getDocumentElement().appendChild(showViewElement);
		
		XMLUtils.appendNewElement(doc, showViewElement, "FullAlias", getFullAlias());
		showViewElement.appendChild(flow.toXML(doc));
		
		FlowFamilyDeduplicationSettings settings = flowFamilyDeduplicationSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());

		boolean enabled = false;
		
		if(settings != null) {
		
			XMLUtils.append(doc, showViewElement, settings);
			enabled = true;
		}
		
		return new FlowAdminExtensionShowView(viewFragmentTransformer.createViewFragment(doc, true), enabled);
	}
	
	@Override
	public String getExtensionViewTitle() {
		
		return extensionViewTitle;
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse updateSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);
		
		if (flow == null) {
			
			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));
			
		} else if (!flowAdminModule.hasFlowAccess(user, flow)) {
			
			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}
		
		FlowFamilyDeduplicationSettings settings = flowFamilyDeduplicationSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());
		
		List<ValidationError> validationErrors = null;
		
		if (req.getMethod().equalsIgnoreCase("POST")) {
			
			validationErrors = new ArrayList<ValidationError>();
			
			if (settings == null) {
				
				settings = new FlowFamilyDeduplicationSettings();
			}
			
			try {
				POPULATOR.populate(settings, req);
				
				String mode = ValidationUtils.validateParameter("mode", req, true, 0, 255, validationErrors);
				
				if (mode != null) {
					
					if ("poster".equals(mode)) {
						
						settings.setMatchPoster(true);
						settings.setAttribute(null);
						
					} else if ("both".equals(mode)) {
						
						settings.setMatchPoster(true);
						
						if (StringUtils.isEmpty(settings.getAttribute())) {
							
							validationErrors.add(new ValidationError("attribute", ValidationErrorType.RequiredField));
						}
						
					} else if ("attribute".equals(mode)) {
						
						settings.setMatchPoster(false);
						
						if (StringUtils.isEmpty(settings.getAttribute())) {
							
							validationErrors.add(new ValidationError("attribute", ValidationErrorType.RequiredField));
						}
						
					} else {
						
						validationErrors.add(new ValidationError("mode", ValidationErrorType.InvalidFormat));
					}
				}
				
				if (!StringUtils.isEmpty(settings.getStatusName()) && getStatus(flow, settings.getStatusName()) == null) {
					
					validationErrors.add(new ValidationError("statusName", ValidationErrorType.InvalidFormat));
				}
				
				if (CollectionUtils.isEmpty(validationErrors)) {
						
					log.info("User " + user + " updated deduplication settings for flow " + flow);
					
					if (settings.getFlowFamilyID() == null) {
						
						settings.setFlowFamilyID(flow.getFlowFamily().getFlowFamilyID());
						flowFamilyDeduplicationSettingsDAOWrapper.add(settings);
						
					} else {
						
						flowFamilyDeduplicationSettingsDAOWrapper.update(settings);
					}
					
					flowAdminModule.addFlowFamilyEvent(flowAdminModule.getEventFunctionConfigured() + " " + flowEventDescription, flow.getFlowFamily(), user);
					flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#deduplicationsettings");
					return null;
				}
				
			} catch (ValidationException e) {
				
				validationErrors.addAll(e.getErrors());
			}
		}
		
		log.info("User " + user + " requesting update deduplication settings form for flow " + flow);
		
		Document doc = createDocument(req, uriParser);
		
		Element settingsElement = doc.createElement("UpdateSettings");
		doc.getDocumentElement().appendChild(settingsElement);
		
		settingsElement.appendChild(flow.toXML(doc));
		
		XMLUtils.append(doc, settingsElement, settings);
		
		if (validationErrors != null) {
			
			XMLUtils.append(doc, settingsElement, "ValidationErrors", validationErrors);
			settingsElement.appendChild(RequestUtils.getRequestParameters(req, doc));
		}
		
		return new SimpleForegroundModuleResponse(doc, this.getDefaultBreadcrumb());
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse deleteSettings(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		if(!HTTPUtils.isPost(req)) {
			
			throw new AccessDeniedException("Delete requests using method " + req.getMethod() + " are not allowed.");
		}
		
		Flow flow = flowAdminModule.getRequestedFlow(req, user, uriParser);
		
		if (flow == null) {
			
			return flowAdminModule.list(req, res, user, uriParser, new ValidationError("FlowNotFound"));
			
		} else if (!flowAdminModule.hasFlowAccess(user, flow)) {
			
			throw new AccessDeniedException("User does not have access to flow type " + flow.getFlowType());
		}
		
		FlowFamilyDeduplicationSettings settings = flowFamilyDeduplicationSettingsDAOWrapper.get(flow.getFlowFamily().getFlowFamilyID());
		
		if (settings != null) {
			
			log.info("User " + user + " deleting deduplication settings for flow " + flow);
			
			flowFamilyDeduplicationSettingsDAOWrapper.delete(settings);
			
		} else {
			
			log.warn("User " + user + " trying to delete deduplication settings for flow " + flow + " which has no settings");
		}
		
		flowAdminModule.redirectToMethod(req, res, "/showflow/" + flow.getFlowID() + "#deduplicationsettings");
		
		return null;
	}
	
	@Override
	public int getModuleID() {

		return moduleDescriptor.getModuleID();
	}
	
	@EventListener(channel = FlowFamily.class)
	public void processFlowFamilyEvent(CRUDEvent<FlowFamily> event, EventSource eventSource) {

		if (event.getAction() == CRUDAction.DELETE) {

			for (FlowFamily flowFamily : event.getBeans()) {

				log.info("Deleting deduplication settings for flow family ID: " + flowFamily.getFlowFamilyID());

				try {
					flowFamilyDeduplicationSettingsDAOWrapper.deleteByID(flowFamily.getFlowFamilyID());

				} catch (SQLException e) {

					log.error("Error deleting deduplication settings for flow family ID: " + flowFamily.getFlowFamilyID(), e);
				}
			}
		}
	}
}
