package com.nordicpeak.flowengine.flowinstanceadminremoteinfo;

import java.io.IOException;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ViewFragment;
import se.unlogic.hierarchy.core.utils.AttributeTagUtils;
import se.unlogic.openhierarchy.foregroundmodules.siteprofile.interfaces.SiteProfile;
import se.unlogic.standardutils.datatypes.SimpleEntry;
import se.unlogic.standardutils.string.AnnotatedBeanTagSourceFactory;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.standardutils.string.URLEncodingTagReplacer;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPResponse;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.Flow;
import com.nordicpeak.flowengine.beans.FlowFamily;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.beans.Status;
import com.nordicpeak.flowengine.flowinstanceadminremoteinfo.beans.FlowInstanceAdminRemoteInfoEndpoint;
import com.nordicpeak.flowengine.flowinstanceadminremoteinfo.beans.FlowInstanceAdminRemoteInfoSetting;
import com.nordicpeak.flowengine.interfaces.FlowInstanceOverviewDescriptionExtensionProvider;
import com.nordicpeak.flowengine.utils.TextTagReplacer;
import com.nordicpeak.flowengine.utils.UserAttributeTagUtils;

public class FlowInstanceAdminRemoteInfoFragment implements FlowInstanceOverviewDescriptionExtensionProvider {

	protected Logger log = Logger.getLogger(this.getClass());

	private static final AnnotatedBeanTagSourceFactory<FlowFamily> FLOW_FAMILY_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(FlowFamily.class, "$flowFamily.");
	private static final AnnotatedBeanTagSourceFactory<Flow> FLOW_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(Flow.class, "$flow.");
	private static final AnnotatedBeanTagSourceFactory<FlowInstance> FLOWINSTANCE_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(FlowInstance.class, "$flowInstance.");
	private static final AnnotatedBeanTagSourceFactory<Status> STATUS_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(Status.class, "$status.");
	private static final AnnotatedBeanTagSourceFactory<User> USER_TAG_SOURCE_FACTORY = new AnnotatedBeanTagSourceFactory<>(User.class, "$user.");

	private final FlowInstanceAdminRemoteInfoModule callback;
	private Integer connectionTimeout;
	private Integer readTimeout;

	public FlowInstanceAdminRemoteInfoFragment(FlowInstanceAdminRemoteInfoModule callback, Integer connectionTimeout, Integer readTimeout) {

		this.callback = callback;
		this.connectionTimeout = connectionTimeout;
		this.readTimeout = readTimeout;
	}

	@Override
	public ViewFragment getFlowInstanceOverviewDescriptionViewFragment(FlowInstance flowInstance, HttpServletRequest req, URIParser uriParser, User user) throws Exception {

		Document doc = callback.createDocument(req, uriParser, user);

		Element showElement = doc.createElement("ShowRemoteInfo");
		doc.getDocumentElement().appendChild(showElement);

		FlowInstanceAdminRemoteInfoSetting setting = callback.getSettingDAO().get(flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		if (setting != null) {

			SiteProfile profile = callback.getCurrentSiteProfile(req, user, uriParser, flowInstance.getFlow().getFlowFamily());

			String remoteInfo = getRemoteInfo(setting.getEndpoint(), flowInstance, user, profile);

			if (remoteInfo != null) {

				XMLUtils.appendNewElement(doc, showElement, "RemoteInfo", remoteInfo);
			}

			return callback.getViewFragmentTransformer().createViewFragment(doc);

		} else {

			return null;
		}
	}

	private String getRemoteInfo(FlowInstanceAdminRemoteInfoEndpoint endpoint, FlowInstance flowInstance, User user, SiteProfile profile) {

		String address = endpoint.getAddress();

		address = replaceTags(address, flowInstance, user, profile, true);

		SimpleRequest simpleRequest = new SimpleRequest(address);

		simpleRequest.setFollowRedirects(false);
		simpleRequest.setConnectionTimeout((int) Duration.ofSeconds(connectionTimeout).toMillis());
		simpleRequest.setReadTimeout((int) Duration.ofSeconds(readTimeout).toMillis());

		if (endpoint.getUsername() != null && endpoint.getPassword() != null) {

			simpleRequest.setUsername(endpoint.getUsername());
			simpleRequest.setPassword(endpoint.getPassword());
		}

		List<Entry<String, String>> headerEntries = new ArrayList<>(1);

		headerEntries.add(new SimpleEntry<String, String>("NoLoginRedirect", "NoLoginRedirect"));
		simpleRequest.setHeaders(headerEntries);

		HTTPResponse response;
		try {

			response = HTTPUtils.sendHTTPGetRequest(simpleRequest, endpoint.getCharset());

		} catch (IOException e) {

			log.warn("Error getting remote info from endpoint", e);
			return null;
		}

		if (response.getHeaders().get("AdminInfoResponse") == null && getCaseInsensitiveMapValue(response.getHeaders(), "AdminInfoResponse") == null) {

			log.warn("Bad response from " + simpleRequest + " missing AdminInfoReponse header");

			return null;
		}

		return replaceTags(response.getValue(), flowInstance, user, profile, false);
	}

	private String replaceTags(String text, FlowInstance flowInstance, User user, SiteProfile profile, boolean urlEncode) {

		TagReplacer tagReplacer = urlEncode ? new URLEncodingTagReplacer() : new TagReplacer();

		if (flowInstance != null) {

			text = AttributeTagUtils.replaceTags(text, flowInstance.getAttributeHandler(), false, true);

			tagReplacer.addTagSource(FLOW_FAMILY_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getFlow().getFlowFamily()));
			tagReplacer.addTagSource(FLOW_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getFlow()));
			tagReplacer.addTagSource(FLOWINSTANCE_TAG_SOURCE_FACTORY.getTagSource(flowInstance));
			tagReplacer.addTagSource(STATUS_TAG_SOURCE_FACTORY.getTagSource(flowInstance.getStatus()));
		}

		if (user != null) {

			text = UserAttributeTagUtils.replaceTags(text, user);

			tagReplacer.addTagSource(USER_TAG_SOURCE_FACTORY.getTagSource(user));
		}

		text = tagReplacer.replace(text);

		if (profile != null) {

			text = TextTagReplacer.replaceTextTags(text, profile.getSettingHandler());
		}

		return text;
	}

	private static List<String> getCaseInsensitiveMapValue(Map<String,List<String>> map, String key) {
		
		for(Entry<String, List<String>> entry : map.entrySet()) {
			
			if(entry != null && entry.getKey() != null && entry.getKey().equalsIgnoreCase(key)) {
				
				return entry.getValue();
			}
		}
		
		return null;
	}
}
