package com.nordicpeak.flowengine.queries.iconmapquery;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.validation.ValidationException;

import com.nordicpeak.flowengine.beans.InstanceRequestMetadata;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.interfaces.ReservationProvider;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQuery;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQueryInstance;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.IconMapQueryInstanceReservation;
import com.nordicpeak.flowengine.queries.iconmapquery.beans.SelectedRemoteGeometry;
import com.nordicpeak.flowengine.queries.iconmapquery.validationerrors.SelectedRemoteGeometryValidationError;
import com.nordicpeak.flowengine.remotegeometryendpointprovider.RemoteGeometry;

public class IconMapQueryReservationHandler {

	private Logger log = Logger.getLogger(this.getClass());

	private ReservationProvider reservationProvider;

	public IconMapQueryReservationHandler(ReservationProvider reservationProvider) {

		this.reservationProvider = reservationProvider;
	}

	public void validateForReservation(IconMapQueryInstance queryInstance, User poster, InstanceRequestMetadata requestMetadata) throws ValidationException {

		if (poster == null) {

			throw new ValidationException(new ValidationError("RequiresLogin"));
		}

		if (requestMetadata.flowInstanceIsSubmitted() && queryInstance.getReservations() != null) {

			throw new ValidationException(new ValidationError("SubmittedWithReservations"));
		}
	}

	public void setCurrentlyReservedGeometryColors(IconMapQueryInstance queryInstance) {

		List<RemoteGeometry> fetchedGeometries = queryInstance.getFetchedGeometries();

		if (fetchedGeometries == null) {

			return;
		}

		try {

			Integer flowFamilyID = getFlowFamilyID(queryInstance);

			List<String> fetchedObjectIDs = CollectionUtils.map(fetchedGeometries, RemoteGeometry::getObjectID);
			List<String> reservedObjectIDs = reservationProvider.getReservedObjectIDs(fetchedObjectIDs, flowFamilyID, null);

			if (reservedObjectIDs == null) {

				return;
			}

			for (RemoteGeometry geometry : fetchedGeometries) {

				if (reservedObjectIDs.contains(geometry.getObjectID())) {

					geometry.setColor(queryInstance.getQuery().getReservedGeometryColor());
				}
			}

		} catch (SQLException e) {

			log.error("Error getting reserved object IDs", e);
		}
	}

	public void reserveSelectedGeometries(IconMapQueryInstance queryInstance, String flowInstanceManagerID) throws ValidationException {

		List<ValidationError> validationErrors = new ArrayList<>();

		try {

			deleteNoLongerSelectedReservations(queryInstance);

			List<IconMapQueryInstanceReservation> newReservations = createReservations(queryInstance, flowInstanceManagerID, validationErrors);

			queryInstance.setReservations(newReservations);

		} catch (Exception e) {

			log.error("Error reserving selected geometries for queryInstance " + queryInstance, e);
			throw new ValidationException(new ValidationError("ReservationError"));
		}

		if (!validationErrors.isEmpty()) {

			throw new ValidationException(validationErrors);
		}
	}

	public void deleteNoLongerSelectedReservations(IconMapQueryInstance queryInstance) throws Exception {

		List<IconMapQueryInstanceReservation> reservations = queryInstance.getReservations();

		if (reservations == null) {

			return;
		}

		Integer flowFamilyID = getFlowFamilyID(queryInstance);

		List<String> reservationIDsToDelete = null;

		for (Iterator<IconMapQueryInstanceReservation> it = reservations.iterator(); it.hasNext();) {

			IconMapQueryInstanceReservation reservation = it.next();

			boolean geometryIsNoLongerSelected = CollectionUtils.find(queryInstance.getSelectedGeometries(), geometry -> reservation.getObjectID().equals(geometry.getObjectID())) == null;

			if (geometryIsNoLongerSelected) {

				reservationIDsToDelete = CollectionUtils.addAndInstantiateIfNeeded(reservationIDsToDelete, reservation.getReservationID());

				it.remove();
			}
		}

		if (reservationIDsToDelete != null) {

			reservationProvider.deleteReservations(reservationIDsToDelete, flowFamilyID);
		}
	}

	public void deleteReservation(String objectID, IconMapQueryInstance queryInstance) throws Exception {

		IconMapQueryInstanceReservation reservation = CollectionUtils.find(queryInstance.getReservations(), r -> r.getObjectID().equals(objectID));

		if (reservation != null) {

			Integer flowFamilyID = getFlowFamilyID(queryInstance);

			reservationProvider.deleteReservation(reservation.getReservationID(), flowFamilyID);

			queryInstance.getReservations().remove(reservation);
		}
	}

	private List<IconMapQueryInstanceReservation> createReservations(IconMapQueryInstance queryInstance, String flowInstanceManagerID, List<ValidationError> validationErrors) throws SQLException {

		List<SelectedRemoteGeometry> geometries = queryInstance.getSelectedGeometries();

		if (geometries == null) {

			return null;
		}

		IconMapQuery query = queryInstance.getQuery();

		List<IconMapQueryInstanceReservation> newReservations = null;

		for (SelectedRemoteGeometry geometry : geometries) {

			IconMapQueryInstanceReservation reservation = createReservation(geometry.getObjectID(), queryInstance, flowInstanceManagerID);

			if (reservation == null) {

				validationErrors = CollectionUtils.addAndInstantiateIfNeeded(validationErrors, new SelectedRemoteGeometryValidationError(geometry, "ReservationFailed", query.getGeometryAlreadyReservedMessage()));
				continue;
			}

			newReservations = CollectionUtils.addAndInstantiateIfNeeded(newReservations, reservation);
		}

		return newReservations;
	}

	public IconMapQueryInstanceReservation createReservation(String objectID, IconMapQueryInstance queryInstance, String flowInstanceManagerID) throws SQLException {

		IconMapQuery query = queryInstance.getQuery();
		Integer reservationMinutes = query.getReservationMinutes();
		Integer flowFamilyID = getFlowFamilyID(queryInstance);
		Integer flowInstanceID = getFlowInstanceID(queryInstance);

		IconMapQueryInstanceReservation reservation = CollectionUtils.find(queryInstance.getReservations(), r -> r.getObjectID().equals(objectID));

		String oldReservationID = reservation != null ? reservation.getReservationID() : null;

		String reservationID = reservationProvider.createReservation(objectID, oldReservationID, flowFamilyID, flowInstanceID, flowInstanceManagerID, reservationMinutes);

		if (reservationID == null) {

			return null;
		}

		if (reservation == null) {

			reservation = new IconMapQueryInstanceReservation(objectID, reservationID);

		} else {

			reservation.setReservationID(reservationID);
		}

		return reservation;
	}

	private Integer getFlowFamilyID(IconMapQueryInstance queryInstance) {

		return queryInstance.getQueryInstanceDescriptor().getQueryDescriptor().getStep().getFlow().getFlowFamily().getFlowFamilyID();
	}

	private Integer getFlowInstanceID(QueryInstance queryInstance) {

		return queryInstance.getQueryInstanceDescriptor().getFlowInstanceID();
	}

}
