package com.nordicpeak.lmsearch;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.CoordinateSequence;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

import net.opengis.gml._3.AbstractGeometryType;
import net.opengis.gml._3.AbstractRingPropertyType;
import net.opengis.gml._3.BoundingShapeType;
import net.opengis.gml._3.CurvePropertyType;
import net.opengis.gml._3.DirectPositionListType;
import net.opengis.gml._3.DirectPositionType;
import net.opengis.gml._3.EnvelopeType;
import net.opengis.gml._3.LineStringType;
import net.opengis.gml._3.LinearRingType;
import net.opengis.gml._3.MultiCurveType;
import net.opengis.gml._3.MultiPointType;
import net.opengis.gml._3.MultiSurfaceType;
import net.opengis.gml._3.ObjectFactory;
import net.opengis.gml._3.PointPropertyType;
import net.opengis.gml._3.PointType;
import net.opengis.gml._3.PolygonType;
import net.opengis.gml._3.SurfacePropertyType;

public class GMLConverter {

	public static final ObjectFactory GML_OBJECT_FACTORY = new ObjectFactory();
	private static final GeometryFactory JTS_GEOMETRY_FACTORY = new GeometryFactory();
	private static final List<Integer> DEFAULT_REFERENCE_SYSTEMS;
	
	static{
		List<Integer> neRefSys = new ArrayList<Integer>();
		
		neRefSys.add(3006);
		neRefSys.add(3007);
		neRefSys.add(3008);
		neRefSys.add(3009);
		neRefSys.add(3010);
		neRefSys.add(3011);
		neRefSys.add(3012);
		neRefSys.add(3013);
		neRefSys.add(3014);
		neRefSys.add(3015);
		neRefSys.add(3016);
		neRefSys.add(3017);
		neRefSys.add(3018);
		neRefSys.add(0);
		
		DEFAULT_REFERENCE_SYSTEMS = Collections.unmodifiableList(neRefSys);
	}
	
	private final AtomicInteger counter = new AtomicInteger();
	
	private boolean swapAxes = false;
	
	private List<Integer> neReferenceSystems;

	public GMLConverter() {

		this.neReferenceSystems = DEFAULT_REFERENCE_SYSTEMS;
	}

	public GMLConverter(boolean swapAxes) {

		this();
		this.swapAxes = swapAxes;
		this.neReferenceSystems = DEFAULT_REFERENCE_SYSTEMS;
	}

	public void setNeReferenceSystems(List<Integer> neReferenceSystems) {

		this.neReferenceSystems = Collections.unmodifiableList(neReferenceSystems);
	}

	public void setSwapAxes(boolean swapAxes) {

		this.swapAxes = swapAxes;
	}

	public Geometry toJtsGeometry(AbstractGeometryType vendorGeometry) {

		return fromAbstractGeometryTypeInternal(vendorGeometry);
	}

	public AbstractGeometryType toVendorGeometry(Geometry jtsGeometry) {

		return toAbstractGeometryTypeInternal(jtsGeometry);
	}

	public static ObjectFactory getGmlObjectFactory() {

		return GML_OBJECT_FACTORY;
	}

	protected BoundingShapeType toBoundingShapeTypeInternal(Geometry geometry) {

		BoundingShapeType boundingShapeType = GML_OBJECT_FACTORY.createBoundingShapeType();
		EnvelopeType envelopeType = GML_OBJECT_FACTORY.createEnvelopeType();
		envelopeType.setSrsName("EPSG:" + geometry.getSRID());
		Envelope envelope = geometry.getEnvelopeInternal();
		envelopeType.setLowerCorner(toDirectPositionTypeInternal(envelope.getMinX(), envelope.getMinY(), geometry.getSRID()));
		envelopeType.setUpperCorner(toDirectPositionTypeInternal(envelope.getMaxX(), envelope.getMaxY(), geometry.getSRID()));
		boundingShapeType.setEnvelope(envelopeType);
		return boundingShapeType;
	}

	protected AbstractGeometryType toAbstractGeometryTypeInternal(Geometry geometry) {

		if (geometry.getClass() == Point.class) {
			return toPointTypeInternal((Point) geometry);
		}
		if (geometry.getClass() == LineString.class) {
			return toLineStringTypeInternal((LineString) geometry);
		}
		if (geometry.getClass() == Polygon.class) {
			return toPolygonTypeInternal((Polygon) geometry);
		}
		if (geometry.getClass() == MultiPoint.class) {
			return toMultiPointTypeInternal((MultiPoint) geometry);
		}
		if (geometry.getClass() == MultiLineString.class) {
			return toMultiCurveTypeInternal((MultiLineString) geometry);
		}
		if (geometry.getClass() == MultiPolygon.class) {
			return toMultiSurfaceTypeInternal((MultiPolygon) geometry);
		}
		throw new IllegalArgumentException("Unsupported geometry: " + geometry);
	}

	public static Envelope fromBoundingShapeType(BoundingShapeType boundingShapeType) {

		EnvelopeType envelopeType = boundingShapeType.getEnvelope();
		List<Double> lowerCorner = envelopeType.getLowerCorner().getValue();
		List<Double> upperCorner = envelopeType.getUpperCorner().getValue();
		Envelope envelope = new Envelope();
		envelope.expandToInclude(lowerCorner.get(0), lowerCorner.get(1));
		envelope.expandToInclude(upperCorner.get(0), upperCorner.get(1));
		return envelope;
	}

	protected Geometry fromAbstractGeometryTypeInternal(AbstractGeometryType geometryType) {

		if (geometryType.getClass() == PointType.class) {
			return fromPointTypeInternal((PointType) geometryType);
		}
		if (geometryType.getClass() == LineStringType.class) {
			return fromLineStringTypeInternal((LineStringType) geometryType);
		}
		if (geometryType.getClass() == PolygonType.class) {
			return fromPolygonTypeInternal((PolygonType) geometryType);
		}
		if (geometryType.getClass() == MultiPointType.class) {
			return fromMultiPointTypeInternal((MultiPointType) geometryType);
		}
		if (geometryType.getClass() == MultiCurveType.class) {
			return fromMultiCurveTypeInternal((MultiCurveType) geometryType);
		}
		if (geometryType.getClass() == MultiSurfaceType.class) {
			return fromMultiSurfaceTypeInternal((MultiSurfaceType) geometryType);
		}
		throw new IllegalArgumentException("Unsupported geometry: " + geometryType);
	}

	protected PointType toPointTypeInternal(Point point) {

		PointType pointType = GML_OBJECT_FACTORY.createPointType();
		pointType.setId(getId());
		pointType.setSrsName("EPSG:" + point.getSRID());
		if (!isTwoDimensional(point)) {
			pointType.setSrsDimension(getDimension(point));
		}
		pointType.setPos(toDirectPositionTypeInternal(point.getCoordinate(), point.getSRID()));
		return pointType;
	}

	protected LineStringType toLineStringTypeInternal(LineString lineString) {

		LineStringType lineStringType = GML_OBJECT_FACTORY.createLineStringType();
		lineStringType.setId(getId());
		lineStringType.setSrsName("EPSG:" + lineString.getSRID());
		if (!isTwoDimensional(lineString)) {
			lineStringType.setSrsDimension(getDimension(lineString));
		}
		lineStringType.setPosList(toDirectPositionListTypeInternal(lineString.getCoordinates(), lineString.getSRID()));
		return lineStringType;
	}

	protected PolygonType toPolygonTypeInternal(Polygon polygon) {

		PolygonType polygonType = GML_OBJECT_FACTORY.createPolygonType();
		polygonType.setId(getId());
		polygonType.setSrsName("EPSG:" + polygon.getSRID());
		if (!isTwoDimensional(polygon)) {
			polygonType.setSrsDimension(getDimension(polygon));
		}
		AbstractRingPropertyType exteriorRing = GML_OBJECT_FACTORY.createAbstractRingPropertyType();
		exteriorRing.setAbstractRing(GML_OBJECT_FACTORY.createLinearRing(toLinearRingTypeInternal(polygon.getExteriorRing())));
		polygonType.setExterior(exteriorRing);
		List<AbstractRingPropertyType> interiorList = polygonType.getInterior();
		final int interiorRingCount = polygon.getNumInteriorRing();
		for (int i = 0; i < interiorRingCount; i++) {
			AbstractRingPropertyType interiorRing = GML_OBJECT_FACTORY.createAbstractRingPropertyType();
			interiorRing.setAbstractRing(GML_OBJECT_FACTORY.createLinearRing(toLinearRingTypeInternal(polygon.getInteriorRingN(i))));
			interiorList.add(interiorRing);
		}
		return polygonType;
	}

	protected MultiPointType toMultiPointTypeInternal(MultiPoint multiPoint) {

		MultiPointType multiPointType = new MultiPointType();
		multiPointType.setId(getId());
		multiPointType.setSrsName("EPSG:" + multiPoint.getSRID());
		if (!isTwoDimensional(multiPoint)) {
			multiPointType.setSrsDimension(getDimension(multiPoint));
		}
		List<PointPropertyType> pointMemberList = multiPointType.getPointMember();
		final int pointCount = multiPoint.getNumGeometries();
		for (int i = 0; i < pointCount; i++) {
			PointPropertyType pointMember = GML_OBJECT_FACTORY.createPointPropertyType();
			PointType pointType = toPointTypeInternal((Point) multiPoint.getGeometryN(i));
			if (multiPointType.getSrsName().equals(pointType.getSrsName())) {
				pointType.setSrsName(null);
			}
			if (objectsEquals(multiPointType.getSrsDimension(), pointType.getSrsDimension())) {
				pointType.setSrsDimension(null);
			} else {
				pointType.setSrsDimension(getDimension(multiPoint.getGeometryN(i)));
			}
			pointMember.setPoint(pointType);
			pointMemberList.add(pointMember);
		}
		return multiPointType;
	}

	protected MultiCurveType toMultiCurveTypeInternal(MultiLineString multiLineString) {

		MultiCurveType multiCurveType = GML_OBJECT_FACTORY.createMultiCurveType();
		multiCurveType.setId(getId());
		multiCurveType.setSrsName("EPSG:" + multiLineString.getSRID());
		if (!isTwoDimensional(multiLineString)) {
			multiCurveType.setSrsDimension(getDimension(multiLineString));
		}
		List<CurvePropertyType> curveMemberList = multiCurveType.getCurveMember();
		final int curveCount = multiLineString.getNumGeometries();
		for (int i = 0; i < curveCount; i++) {
			CurvePropertyType curveMember = GML_OBJECT_FACTORY.createCurvePropertyType();
			LineStringType lineStringType = toLineStringTypeInternal((LineString) multiLineString.getGeometryN(i));
			if (multiCurveType.getSrsName().equals(lineStringType.getSrsName())) {
				lineStringType.setSrsName(null);
			}
			if (objectsEquals(multiCurveType.getSrsDimension(), lineStringType.getSrsDimension())) {
				lineStringType.setSrsDimension(null);
			} else {
				lineStringType.setSrsDimension(getDimension(multiLineString.getGeometryN(i)));
			}
			curveMember.setAbstractCurve(GML_OBJECT_FACTORY.createLineString(lineStringType));
			curveMemberList.add(curveMember);
		}
		return multiCurveType;
	}

	protected MultiSurfaceType toMultiSurfaceTypeInternal(MultiPolygon multiPolygon) {

		MultiSurfaceType multiSurfaceType = GML_OBJECT_FACTORY.createMultiSurfaceType();
		multiSurfaceType.setId(getId());
		multiSurfaceType.setSrsName("EPSG:" + multiPolygon.getSRID());
		if (!isTwoDimensional(multiPolygon)) {
			multiSurfaceType.setSrsDimension(getDimension(multiPolygon));
		}
		List<SurfacePropertyType> surfaceMemberList = multiSurfaceType.getSurfaceMember();
		final int surfaceCount = multiPolygon.getNumGeometries();
		for (int i = 0; i < surfaceCount; i++) {
			SurfacePropertyType surfaceMember = GML_OBJECT_FACTORY.createSurfacePropertyType();
			PolygonType polygonType = toPolygonTypeInternal((Polygon) multiPolygon.getGeometryN(i));
			if (multiSurfaceType.getSrsName().equals(polygonType.getSrsName())) {
				polygonType.setSrsName(null);
			}
			if (objectsEquals(multiSurfaceType.getSrsDimension(), polygonType.getSrsDimension())) {
				polygonType.setSrsDimension(null);
			} else {
				polygonType.setSrsDimension(getDimension(multiPolygon.getGeometryN(i)));
			}
			surfaceMember.setAbstractSurface(GML_OBJECT_FACTORY.createPolygon(polygonType));
			surfaceMemberList.add(surfaceMember);
		}
		return multiSurfaceType;
	}

	protected Point fromPointTypeInternal(PointType pointType) {

		int srid = extractSrid(pointType.getSrsName());
		Point point = JTS_GEOMETRY_FACTORY.createPoint(toCoordinateSequenceInternal(pointType.getPos().getValue(), pointType.getSrsDimension(), srid));
		point.setSRID(srid);
		return point;
	}

	protected Point fromPointTypeInternal(PointType pointType, AbstractGeometryType parent) {

		int srid = pointType.getSrsName() != null ? extractSrid(pointType.getSrsName()) : extractSrid(parent.getSrsName());
		BigInteger dimension = pointType.getSrsDimension() != null ? pointType.getSrsDimension() : parent.getSrsDimension();
		Point point = JTS_GEOMETRY_FACTORY.createPoint(toCoordinateSequenceInternal(pointType.getPos().getValue(), dimension, srid));
		point.setSRID(srid);
		return point;
	}

	protected LineString fromLineStringTypeInternal(LineStringType lineStringType) {

		int srid = extractSrid(lineStringType.getSrsName());
		LineString lineString = JTS_GEOMETRY_FACTORY.createLineString(toCoordinateSequenceInternal(lineStringType.getPosList().getValue(), lineStringType.getSrsDimension(), srid));
		lineString.setSRID(srid);
		return lineString;
	}

	protected LineString fromLineStringTypeInternal(LineStringType lineStringType, AbstractGeometryType parent) {

		int srid = lineStringType.getSrsName() != null ? extractSrid(lineStringType.getSrsName()) : extractSrid(parent.getSrsName());
		BigInteger dimension = lineStringType.getSrsDimension() != null ? lineStringType.getSrsDimension() : parent.getSrsDimension();
		LineString lineString = JTS_GEOMETRY_FACTORY.createLineString(toCoordinateSequenceInternal(lineStringType.getPosList().getValue(), dimension, srid));
		lineString.setSRID(srid);
		return lineString;
	}

	protected Polygon fromPolygonTypeInternal(PolygonType polygonType) {

		int srid = extractSrid(polygonType.getSrsName());
		LinearRingType exteriorRing = (LinearRingType) polygonType.getExterior().getAbstractRing().getValue();
		LinearRing exterior = JTS_GEOMETRY_FACTORY.createLinearRing(toCoordinateSequenceInternal(exteriorRing.getPosList().getValue(), polygonType.getSrsDimension(), srid));
		LinearRing[] interiorArray = new LinearRing[polygonType.getInterior().size()];
		for (int i = 0; i < interiorArray.length; i++) {
			LinearRingType interiorRing = (LinearRingType) polygonType.getInterior().get(i).getAbstractRing().getValue();
			interiorArray[i] = JTS_GEOMETRY_FACTORY.createLinearRing(toCoordinateSequenceInternal(interiorRing.getPosList().getValue(), polygonType.getSrsDimension(), srid));
		}
		Polygon polygon = JTS_GEOMETRY_FACTORY.createPolygon(exterior, interiorArray);
		polygon.setSRID(srid);
		return polygon;
	}

	protected Polygon fromPolygonTypeInternal(PolygonType polygonType, AbstractGeometryType parent) {

		int srid = polygonType.getSrsName() != null ? extractSrid(polygonType.getSrsName()) : extractSrid(parent.getSrsName());
		BigInteger dimension = polygonType.getSrsDimension() != null ? polygonType.getSrsDimension() : parent.getSrsDimension();
		LinearRingType exteriorRing = (LinearRingType) polygonType.getExterior().getAbstractRing().getValue();
		LinearRing exterior = JTS_GEOMETRY_FACTORY.createLinearRing(toCoordinateSequenceInternal(exteriorRing.getPosList().getValue(), dimension, srid));
		LinearRing[] interiorArray = new LinearRing[polygonType.getInterior().size()];
		for (int i = 0; i < interiorArray.length; i++) {
			LinearRingType interiorRing = (LinearRingType) polygonType.getInterior().get(i).getAbstractRing().getValue();
			interiorArray[i] = JTS_GEOMETRY_FACTORY.createLinearRing(toCoordinateSequenceInternal(interiorRing.getPosList().getValue(), dimension, srid));
		}
		Polygon polygon = JTS_GEOMETRY_FACTORY.createPolygon(exterior, interiorArray);
		polygon.setSRID(srid);
		return polygon;
	}

	protected MultiPoint fromMultiPointTypeInternal(MultiPointType multiPointType) {

		List<PointPropertyType> pointMember = multiPointType.getPointMember();
		Point[] pointArray = new Point[pointMember.size()];
		for (int i = 0; i < pointArray.length; i++) {
			pointArray[i] = fromPointTypeInternal(pointMember.get(i).getPoint(), multiPointType);
		}
		MultiPoint multiPoint = JTS_GEOMETRY_FACTORY.createMultiPoint(pointArray);
		multiPoint.setSRID(extractSrid(multiPointType.getSrsName()));
		return multiPoint;
	}

	protected MultiLineString fromMultiCurveTypeInternal(MultiCurveType multiCurveType) {

		List<CurvePropertyType> curveMember = multiCurveType.getCurveMember();
		LineString[] lineStringArray = new LineString[curveMember.size()];
		for (int i = 0; i < lineStringArray.length; i++) {
			LineStringType member = (LineStringType) curveMember.get(i).getAbstractCurve().getValue();
			lineStringArray[i] = fromLineStringTypeInternal(member, multiCurveType);
		}
		MultiLineString multiLineString = JTS_GEOMETRY_FACTORY.createMultiLineString(lineStringArray);
		multiLineString.setSRID(extractSrid(multiCurveType.getSrsName()));
		return multiLineString;
	}

	protected MultiPolygon fromMultiSurfaceTypeInternal(MultiSurfaceType multiSurfaceType) {

		List<SurfacePropertyType> surfaceMember = multiSurfaceType.getSurfaceMember();
		Polygon[] polygonArray = new Polygon[surfaceMember.size()];
		for (int i = 0; i < polygonArray.length; i++) {
			PolygonType member = (PolygonType) surfaceMember.get(i).getAbstractSurface().getValue();
			polygonArray[i] = fromPolygonTypeInternal(member, multiSurfaceType);
		}
		MultiPolygon multiPolygon = JTS_GEOMETRY_FACTORY.createMultiPolygon(polygonArray);
		multiPolygon.setSRID(extractSrid(multiSurfaceType.getSrsName()));
		return multiPolygon;
	}

	protected DirectPositionType toDirectPositionTypeInternal(double x, double y, int srid) {

		DirectPositionType directPositionType = GML_OBJECT_FACTORY.createDirectPositionType();
		List<Double> coordinateList = directPositionType.getValue();
		if (swapAxes && neReferenceSystems.contains(srid)) {
			coordinateList.add(y);
			coordinateList.add(x);
		} else {
			coordinateList.add(x);
			coordinateList.add(y);
		}
		return directPositionType;
	}

	protected DirectPositionType toDirectPositionTypeInternal(Coordinate coordinate, int srid) {

		DirectPositionType directPositionType = GML_OBJECT_FACTORY.createDirectPositionType();
		List<Double> coordinateList = directPositionType.getValue();
		if (swapAxes && neReferenceSystems.contains(srid)) {
			coordinateList.add(coordinate.y);
			coordinateList.add(coordinate.x);
		} else {
			coordinateList.add(coordinate.x);
			coordinateList.add(coordinate.y);
		}
		if (!Double.isNaN(coordinate.z)) {
			coordinateList.add(coordinate.z);
		}
		return directPositionType;
	}

	protected DirectPositionListType toDirectPositionListTypeInternal(Coordinate[] coordinates, int srid) {

		DirectPositionListType directPositionListType = GML_OBJECT_FACTORY.createDirectPositionListType();
		List<Double> coordinateList = directPositionListType.getValue();
		for (Coordinate coordinate : coordinates) {
			if (swapAxes && neReferenceSystems.contains(srid)) {
				coordinateList.add(coordinate.y);
				coordinateList.add(coordinate.x);
			} else {
				coordinateList.add(coordinate.x);
				coordinateList.add(coordinate.y);
			}
			if (!Double.isNaN(coordinate.z)) {
				coordinateList.add(coordinate.z);
			}
		}
		return directPositionListType;
	}

	protected LinearRingType toLinearRingTypeInternal(LineString lineString) {

		LinearRingType linearRingType = GML_OBJECT_FACTORY.createLinearRingType();
		linearRingType.setPosList(toDirectPositionListTypeInternal(lineString.getCoordinates(), lineString.getSRID()));
		return linearRingType;
	}

	protected CoordinateSequence toCoordinateSequenceInternal(List<Double> coordinateList, BigInteger bigDimension, int srid) {

		int dimension = 2;
		if (bigDimension != null) {
			dimension = bigDimension.intValue();
		}
		if (dimension != 2 && dimension != 3) {
			throw new IllegalArgumentException("Invalid dimension: " + dimension + ". Allowed dimensions are 2 and 3.");
		}
		if (coordinateList.size() % dimension != 0) {
			throw new IllegalArgumentException("Invalid number of coordinates found in geometry!");
		}

		// Build list of coordinates.
		Coordinate[] coordinates = new Coordinate[coordinateList.size() / dimension];
		for (int i = 0, j = 0; i < coordinates.length; i++, j += dimension) {
			if (swapAxes && neReferenceSystems.contains(srid)) {
				coordinates[i] = new Coordinate(coordinateList.get(j + 1), coordinateList.get(j));
			} else {
				coordinates[i] = new Coordinate(coordinateList.get(j), coordinateList.get(j + 1));
			}
			if (dimension == 3) {
				coordinates[i].z = coordinateList.get(j + 2);
			}
		}
		return JTS_GEOMETRY_FACTORY.getCoordinateSequenceFactory().create(coordinates);
	}

	protected String getId() {

		return "GM_" + counter.incrementAndGet();
	}

	public static boolean isTwoDimensional(Geometry geometry) {

		return Double.isNaN(geometry.getCoordinate().z);
	}

	public static BigInteger getDimension(Geometry geometry) {

		return Double.isNaN(geometry.getCoordinate().z) ? BigInteger.valueOf(2) : BigInteger.valueOf(3);
	}

	public static boolean objectsEquals(Object a, Object b) {

		return (a == b) || (a != null && a.equals(b));
	}

	public static int extractSrid(String epsg) {

		if (epsg == null) {
			throw new IllegalArgumentException("No srsName found in geometry!");
		}
		return Integer.parseInt(epsg.substring(epsg.lastIndexOf(':') + 1));
	}

}