package com.nordicpeak.origomap;

import java.io.InputStream;
import java.lang.reflect.Field;
import java.util.ArrayDeque;
import java.util.Collection;
import java.util.HashMap;
import java.util.Queue;

import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextAreaSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.reflection.ReflectionUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.validation.StringIntegerValidator;

import com.nordicpeak.lmsearch.LMSearchProvider;
import com.nordicpeak.wmsproxy.WMSProxyModule;

import flexjson.JSONDeserializer;
import flexjson.JSONSerializer;

public class OrigoMapProvider extends AnnotatedRESTModule {

	private static final String PROJECTION_EXTENT_TAG = "$projectionExtent";
	private static final String EXTENT_TAG = "$extent";
	private static final String CENTER_TAG = "$center";
	private static final String ZOOM_TAG = "$zoom";
	private static final String EXTRA_LAYERS_TAG = "$extraLayers";
	private static final String EXTRA_SOURCES_TAG = "$extraSources";
	private static final String STATIC_CONTENT_LINK = "$staticContentLink";
	private static final String LM_SEARCH_PROVIDER_LINK_TAG = "$lmSearchProviderLink";
	private static final String WMS_PROXY_LINK_TAG = "$wmsProxyLink";

	private static final String ORIGO_CONFIG_TEMPLATE = "origoConfig.template";

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "projectionExtent (JSON)", description = "Projection extent as a JSON array: [xmin, ymin, xmax, ymax]")
	private String projectionExtent;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "extent (JSON)", description = "Extent as a JSON array: [xmin, ymin, xmax, ymax]")
	private String extent;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "center (JSON)", description = "Center coordinate as a JSON array: [x, y]")
	private String center;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "zoom (JSON)", description = "Zoom level (Integer)", formatValidator = StringIntegerValidator.class)
	private String zoom;

	@ModuleSetting(allowsNull = true)
	@TextAreaSettingDescriptor(name = "extraLayers (JSON)", description = "Extra layers as one or multiple JSON object(s). Multiple objects must be comma-separated.")
	private String extraLayers;
	
	@ModuleSetting(allowsNull = true)
	@TextAreaSettingDescriptor(name = "extraSources (JSON)", description = "Extra sources as one or multiple JSON object(s). Multiple objects must be comma-separated.")
	private String extraSources;

	private String mapConfiguration;

	private StaticContentModule staticContentModule;

	private WMSProxyModule wmsProxyModule;

	private LMSearchProvider lmSearchProvider;

	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {

		super.init(moduleDescriptor, sectionInterface, dataSource);

		systemInterface.getInstanceHandler().addInstance(OrigoMapProvider.class, this);
	}

	@Override
	public void unload() throws Exception {

		systemInterface.getInstanceHandler().removeInstance(OrigoMapProvider.class, this);

		super.unload();
	}

	@Override
	protected void moduleConfigured() throws Exception {

		super.moduleConfigured();

		try (InputStream inputStream = this.getClass().getResourceAsStream(ORIGO_CONFIG_TEMPLATE)) {

			mapConfiguration = StringUtils.readStreamAsString(inputStream);
		}

		if (mapConfiguration != null) {

			mapConfiguration = replaceTag(mapConfiguration, PROJECTION_EXTENT_TAG, projectionExtent);
			mapConfiguration = replaceTag(mapConfiguration, EXTENT_TAG, extent);
			mapConfiguration = replaceTag(mapConfiguration, CENTER_TAG, center);
			mapConfiguration = replaceTag(mapConfiguration, ZOOM_TAG, zoom);
			
			if (staticContentModule != null) {
				
				mapConfiguration = replaceTag(mapConfiguration, STATIC_CONTENT_LINK, getStaticContentURL());
			}

			if (wmsProxyModule != null) {

				mapConfiguration = replaceTag(mapConfiguration, WMS_PROXY_LINK_TAG, getWMSProxyLink());
			}

			if (lmSearchProvider != null) {

				mapConfiguration = replaceTag(mapConfiguration, LM_SEARCH_PROVIDER_LINK_TAG, getLMSearchProviderLink());
			}

			replaceOptionalMapConfigurationTag(EXTRA_LAYERS_TAG, extraLayers, ",");
			replaceOptionalMapConfigurationTag(EXTRA_SOURCES_TAG, extraSources, ",");
		}
	}

	private String replaceTag(String text, String target, String replacement) {

		if (!StringUtils.isEmpty(replacement)) {

			return text.replace(target, replacement);
		}

		return text;
	}

	private void replaceOptionalMapConfigurationTag(String target, String replacement, String suffix) {

		replacement = replacement != null ? replacement + suffix : "";

		mapConfiguration = mapConfiguration.replace(target, replacement);
	}
	
	@InstanceManagerDependency(required = true)
	public void setStaticContentModule(StaticContentModule staticContentModule) {
		
		this.staticContentModule = staticContentModule;
		
		if (staticContentModule != null && mapConfiguration != null) {
			
			mapConfiguration = replaceTag(mapConfiguration, STATIC_CONTENT_LINK, getStaticContentURL());
		}
	}

	@InstanceManagerDependency(required = true)
	public void setWMSProxyModule(WMSProxyModule wmsProxyModule) {

		this.wmsProxyModule = wmsProxyModule;

		if (wmsProxyModule != null && mapConfiguration != null) {

			mapConfiguration = replaceTag(mapConfiguration, WMS_PROXY_LINK_TAG, getWMSProxyLink());
		}
	}

	@InstanceManagerDependency(required = true)
	public void setLMSearchProvider(LMSearchProvider lmSearchProvider) {

		this.lmSearchProvider = lmSearchProvider;

		if (lmSearchProvider != null && mapConfiguration != null) {

			mapConfiguration = replaceTag(mapConfiguration, LM_SEARCH_PROVIDER_LINK_TAG, getLMSearchProviderLink());
		}
	}

	private String getWMSProxyLink() {

		return systemInterface.getContextPath() + wmsProxyModule.getFullAlias();
	}

	private String getLMSearchProviderLink() {

		return systemInterface.getContextPath() + lmSearchProvider.getFullAlias();
	}

	public String getMapConfiguration() {

		return mapConfiguration;
	}

	public String getStaticContentURL() {

		return staticContentModule.getModuleContentURL(moduleDescriptor);
	}

	public void updateModuleSettings(String mapConfiguration) throws Exception {

		boolean anySettingUpdated = false;

		if (updateModuleSetting(mapConfiguration, "projectionExtent")) {

			anySettingUpdated = true;
		}

		if (updateModuleSetting(mapConfiguration, "extent")) {

			anySettingUpdated = true;
		}

		if (updateModuleSetting(mapConfiguration, "center")) {

			anySettingUpdated = true;
		}

		if (updateModuleSetting(mapConfiguration, "zoom")) {

			anySettingUpdated = true;
		}

		if (anySettingUpdated) {

			this.moduleDescriptor.saveSettings(systemInterface);
			moduleConfigured();
		}
	}

	private boolean updateModuleSetting(String mapConfiguration, String settingName) throws IllegalArgumentException, IllegalAccessException {

		Field setting = ReflectionUtils.getField(this.getClass(), settingName);

		if (setting.get(this) != null) {

			log.info("Module setting " + settingName + " already has a value, not updating");
			return false;
		}

		String valueFromConfig = getSubJSONByKey(mapConfiguration, settingName);

		if (valueFromConfig == null) {

			log.error("Couldn't find configuration for setting " + settingName + " in JSON configuration, module setting need to be manually updated");
			return false;
		}

		log.info("Updating module setting " + settingName + " with value from JSON configuration: " + valueFromConfig);

		this.moduleDescriptor.getMutableSettingHandler().setSetting(settingName, valueFromConfig);
		ReflectionUtils.setFieldValue(setting, valueFromConfig, this);

		return true;
	}

	@SuppressWarnings({ "rawtypes" })
	private static String getSubJSONByKey(String json, String key) {

		JSONSerializer jsonSerializer = new JSONSerializer();
		Object object = new JSONDeserializer<>().deserialize(json);

		Queue<Object> queue = new ArrayDeque<>();

		queue.add(object);

		while (!queue.isEmpty()) {

			Object next = queue.poll();
			Collection collection = null;

			if (next instanceof HashMap) {

				HashMap hashMap = (HashMap) next;

				if (hashMap.containsKey(key)) {

					return jsonSerializer.deepSerialize(hashMap.get(key));
				}

				collection = hashMap.values();

			} else if (next instanceof Collection) {

				collection = (Collection) next;
			}

			if (collection != null) {

				for (Object obj : collection) {

					if (obj instanceof HashMap || obj instanceof Collection) {

						queue.add(obj);
					}
				}
			}
		}

		return null;
	}

}
