package com.nordicpeak.flowengine.paymentlog;

import java.sql.SQLException;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.SectionInterface;
import se.unlogic.hierarchy.core.interfaces.modules.descriptors.ForegroundModuleDescriptor;
import se.unlogic.hierarchy.core.utils.CRUDCallback;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.core.utils.crud.TransactionRequestFilter;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.dao.AdvancedAnnotatedDAOWrapper;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.querys.BooleanQuery;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.enums.EnumUtils;
import se.unlogic.standardutils.validation.ValidationError;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.paymentlog.beans.Account;
import com.nordicpeak.flowengine.paymentlog.beans.AccountNumberPart;
import com.nordicpeak.flowengine.paymentlog.beans.ChartOfAccounts;
import com.nordicpeak.flowengine.paymentlog.cruds.AccountCRUD;
import com.nordicpeak.flowengine.paymentlog.cruds.ChartCRUD;
import com.nordicpeak.flowengine.paymentlog.enums.AccountType;

public class AccountModule extends AnnotatedForegroundModule implements CRUDCallback<User> {
	
	private AdvancedAnnotatedDAOWrapper<Account, Integer> accountDAOWrapper;
	private AdvancedAnnotatedDAOWrapper<ChartOfAccounts, Integer> chartDAOWrapper;
	
	private QueryParameterFactory<Account, AccountType> accountTypeParameterFactory;
	private QueryParameterFactory<Account, ChartOfAccounts> accountChartParameterFactory;
	
	private AccountCRUD accountCRUD;
	private ChartCRUD chartCRUD;
	
	@Override
	public void init(ForegroundModuleDescriptor moduleDescriptor, SectionInterface sectionInterface, DataSource dataSource) throws Exception {
		
		super.init(moduleDescriptor, sectionInterface, dataSource);
		
		if (!systemInterface.getInstanceHandler().addInstance(AccountModule.class, this)) {
			
			throw new RuntimeException("Unable to register module " + this.moduleDescriptor + " in global instance handler using key " + AccountModule.class.getSimpleName() + ", another instance is already registered using this key.");
		}
	}
	
	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {
		
		super.createDAOs(dataSource);
		
		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, AccountModule.class.getName(), new XMLDBScriptProvider(AccountModule.class.getResourceAsStream("DB script.xml")));
		
		if (upgradeResult.isUpgrade()) {
			
			log.info(upgradeResult.toString());
		}
		
		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);
		
		accountDAOWrapper = daoFactory.getDAO(Account.class).getAdvancedWrapper(Integer.class);
		accountDAOWrapper.getGetQuery().addRelations(Account.CHART_RELATION, Account.ACCOUNT_NUMBER_PARTS_RELATION, AccountNumberPart.ACCOUNT_NUMBER_PART_DEFINITION_RELATION);
		accountDAOWrapper.getAddQuery().addRelations(Account.ACCOUNT_NUMBER_PARTS_RELATION);
		accountDAOWrapper.getUpdateQuery().addRelations(Account.ACCOUNT_NUMBER_PARTS_RELATION);
		
		chartDAOWrapper = daoFactory.getDAO(ChartOfAccounts.class).getAdvancedWrapper(Integer.class);
		chartDAOWrapper.getGetQuery().addRelations(ChartOfAccounts.ACCOUNTS_RELATION, Account.ACCOUNT_NUMBER_PARTS_RELATION, AccountNumberPart.ACCOUNT_NUMBER_PART_DEFINITION_RELATION);
		
		accountTypeParameterFactory = accountDAOWrapper.getAnnotatedDAO().getParamFactory("accountType", AccountType.class);
		accountChartParameterFactory = accountDAOWrapper.getAnnotatedDAO().getParamFactory("chart", ChartOfAccounts.class);
		
		accountCRUD = new AccountCRUD(accountDAOWrapper, this);
		accountCRUD.addRequestFilter(new TransactionRequestFilter(dataSource));
		chartCRUD = new ChartCRUD(chartDAOWrapper, this);
	}
	
	@Override
	public void unload() throws Exception {
		
		systemInterface.getInstanceHandler().removeInstance(AccountModule.class, this);
		
		super.unload();
	}
	
	@Override
	public Document createDocument(HttpServletRequest req, URIParser uriParser, User user) {
		
		Document doc = XMLUtils.createDomDocument();
		Element documentElement = doc.createElement("Document");
		documentElement.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		documentElement.appendChild(moduleDescriptor.toXML(doc));
		
//		XMLUtils.appendNewElement(doc, documentElement, "ModuleURI", req.getContextPath() + getFullAlias());
		
		doc.appendChild(documentElement);
		return doc;
	}
	
	@Override
	public String getTitlePrefix() {
		return moduleDescriptor.getName();
	}
	
	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return list(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return chartCRUD.list(req, res, user, uriParser, null);
	}
	
	public ForegroundModuleResponse list(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser, List<ValidationError> validationErrors) throws Exception {
		
		return chartCRUD.list(req, res, user, uriParser, validationErrors);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse showChart(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return chartCRUD.show(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse addChart(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return chartCRUD.add(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse updateChart(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return chartCRUD.update(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse deleteChart(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return chartCRUD.delete(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse showAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return accountCRUD.show(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse addAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		Integer chartID;
		ChartOfAccounts chart;
		String type = req.getParameter("type");
		AccountType accountType;
		
		if (uriParser.size() == 3 && type != null && (chartID = uriParser.getInt(2)) != null && (chart = chartDAOWrapper.get(chartID)) != null && (accountType = EnumUtils.toEnum(AccountType.class, type.toUpperCase())) != null) {
			
			req.setAttribute("chart", chart);
			req.setAttribute("accountType", accountType);
			
			return accountCRUD.add(req, res, user, uriParser);
		}
		
		throw new URINotFoundException(uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse updateAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return accountCRUD.update(req, res, user, uriParser);
	}
	
	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse deleteAccount(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {
		
		return accountCRUD.delete(req, res, user, uriParser);
	}
	
	public List<Account> getAccounts(ChartOfAccounts chart, AccountType type) throws SQLException {
		
		HighLevelQuery<Account> query = new HighLevelQuery<>();
		query.addParameter(accountTypeParameterFactory.getParameter(type));
		query.addParameter(accountChartParameterFactory.getParameter(chart));
		
		return accountDAOWrapper.getAnnotatedDAO().getAll(query);
	}

	public boolean isAccountInUse(Account account) throws SQLException {
		
		//TODO also check if any PaymentEvent listener is using the account
		
		BooleanQuery query = new BooleanQuery(dataSource, "SELECT transactionID FROM payment_transactions WHERE accountID = ? LIMIT 1");
		query.setInt(1, account.getAccountID());
		return query.executeQuery();
	}

	public boolean isChartInUse(ChartOfAccounts chart) throws SQLException {
		
		//TODO also check if any PaymentEvent listener is using any of the charts accounts
		
		BooleanQuery query = new BooleanQuery(dataSource, "SELECT t.transactionID FROM payment_transactions t INNER JOIN payment_accounts a ON t.accountID = a.accountID WHERE a.chartID = ? LIMIT 1");
		query.setInt(1, chart.getChartID());
		return query.executeQuery();
	}
	
}
