package com.nordicpeak.flowengine.queries.personquery;

import java.io.IOException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.populators.SwedishSocialSecurity12DigitsWithoutMinusPopulator;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.interfaces.QueryRequestProcessor;
import com.nordicpeak.flowengine.utils.CitizenIdentifierUtils;
import com.nordicpeak.personprovider.Person;
import com.nordicpeak.personprovider.PersonProvider;

public class PersonQueryRequestProcessor implements QueryRequestProcessor {

	protected final Logger log = Logger.getLogger(this.getClass());

	private final PersonProvider personProvider;
	private final PersonQueryInstance queryInstance;
	private final int maxLookups;
	private final User poster;

	public PersonQueryRequestProcessor(PersonProvider personProvider, PersonQueryInstance queryInstance, int maxLookups, User poster) {

		super();

		this.personProvider = personProvider;
		this.queryInstance = queryInstance;
		this.maxLookups = maxLookups;
		this.poster = poster;
	}

	@Override
	public void processRequest(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		String citizenID = req.getParameter("q");

		if (StringUtils.isEmpty(citizenID) || !SwedishSocialSecurity12DigitsWithoutMinusPopulator.getPopulator().validateFormat(citizenID)) {

			sendErrorJSONResponse(res, "invalid format");
			return;
		}

		String posterCitizenID = CitizenIdentifierUtils.getUserOrManagerCitizenIdentifier(poster);

		if (queryInstance.getQuery().isBlockPosterSelection() && citizenID.equals(posterCitizenID)) {

			sendErrorJSONResponse(res, "block poster selection");
			return;
		}

		if (queryInstance.getCache().containsKey(citizenID)) {

			Person person = queryInstance.getCache().get(citizenID);
			sendJSONResponse(person, res);
			return;
		}

		if (queryInstance.getCache().size() >= maxLookups) {

			sendErrorJSONResponse(res, "max lookups reached");
			return;
		}

		try {

			if (personProvider == null) {

				log.error("Missing required instance manager dependency: personProvider");
				sendErrorJSONResponse(res, "provider error");
				return;
			}

			Person person = personProvider.getPerson(citizenID);

			if (person != null) {

				sendJSONResponse(person, res);
				queryInstance.getCache().put(citizenID, person);

			} else {

				sendErrorJSONResponse(res, "not found");
			}

		} catch (Exception e) {

			sendErrorJSONResponse(res, "provider error");
			throw e;
		}
	}

	protected static void sendErrorJSONResponse(HttpServletResponse res, String error) throws IOException {

		JsonObject jsonObject = new JsonObject(1);
		jsonObject.putField("error", error);
		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	protected void sendJSONResponse(Person person, HttpServletResponse res) throws IOException {

		JsonObject jsonObject = new JsonObject(2);
		jsonObject.putField("firstname", person.getFirstname());
		jsonObject.putField("lastname", person.getLastname());
		jsonObject.putField("address", person.getAddress());
		jsonObject.putField("postalCode", person.getZipCode());
		jsonObject.putField("postalAddress", person.getPostalAddress());
		jsonObject.putField("careOf", person.getCareOf());
		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
	}

	@Override
	public void close() throws Exception {}

}
