package com.nordicpeak.flowengine.queries.plotquery.clients;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.date.DateUtils;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.string.SingleTagSourceFactory;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.string.TagReplacer;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SessionUtils;
import se.unlogic.webutils.http.SimpleRequest;

import com.nordicpeak.flowengine.queries.plotquery.PlotIntegrationConstants;
import com.nordicpeak.flowengine.queries.plotquery.beans.PlotQueryInstance;
import com.nordicpeak.flowengine.queries.plotquery.clients.tokenhandler.ArcGisTokenHandler;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.InvalidResponseException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotCurrentlyNotBookableException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotLockException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotLockExpiredException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotLockedException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotNotBookableException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotSoldException;
import com.nordicpeak.flowengine.queries.plotquery.exceptions.PlotUnavailableException;

import flexjson.JSONDeserializer;

public class PlotQueryClient {

	public static final Charset INTEGRATION_CHARSET = StandardCharsets.UTF_8;

	private static final String FEATURES = "features";
	private static final String ATTRIBUTES = "attributes";

	private static final String OBJECT_ID = "OBJECTID";
	private static final String PLOT_ID = "TOMTID";
	private static final String PROPERTY_UNIT_DESIGNATION = "FASTIGHET";
	private static final String ADDRESS = "ADRESS";
	private static final String AREA = "AREA_CA";
	private static final String CONNECTION_FEE = "VA";
	private static final String PRICE = "PRIS";
	private static final String LOCATION = "OMRADE";
	private static final String COMMENT = "KOMMENTAR";

	private static final String SOLD = "SALD";
	private static final String BOOKABLE = "BOKNINGSBA";
	private static final String BOOKABLE_DATE = "BOKDATUM";
	private static final String LOCK = "LASNING";

	private static final String TRUE = "1";
	private static final String SOLD_FREE_TEXT = "Ledig";
	private static final String SOLD_BOOKED_TEXT = "Bokad";

	private static final SingleTagSourceFactory<Integer> OBJECT_ID_TAG_SOURCE_FACTORY = new SingleTagSourceFactory<>("$objectID");
	private static final String UPDATE_REQUEST_CONTENT_TYPE = "application/x-www-form-urlencoded";

	private Logger log = Logger.getLogger(this.getClass());

	protected String searchPlotURL;
	protected String updatePlotURL;
	protected int connectionTimeout;
	protected int readTimeout;
	protected Integer lockTime;
	protected String refererHeader;
	protected boolean soldFieldStringMode;
	protected boolean ignoreBookableField;
	private ArcGisTokenHandler tokenHandler;
	protected boolean verboseAPILoggingEnabled;

	public PlotQueryClient(String searchPlotURL, String updatePlotURL, int connectionTimeout, int readTimeout, Integer lockTime, String refererHeader, boolean soldFieldStringMode, boolean ignoreBookableField, ArcGisTokenHandler tokenHandler, boolean verboseAPILoggingEnabled) {

		this.searchPlotURL = searchPlotURL;
		this.updatePlotURL = updatePlotURL;
		this.connectionTimeout = connectionTimeout;
		this.readTimeout = readTimeout;
		this.lockTime = lockTime;
		this.refererHeader = refererHeader;
		this.soldFieldStringMode = soldFieldStringMode;
		this.ignoreBookableField = ignoreBookableField;
		this.tokenHandler = tokenHandler;
		this.verboseAPILoggingEnabled = verboseAPILoggingEnabled;
	}

	public void setPlotAsSold(PlotQueryInstance queryInstance, String loggingPrefix) {

		log.info(loggingPrefix + "Setting plot as sold for " + queryInstance);

		Map<String, String> attributes = new HashMap<>();

		attributes.put(LOCK, null);
		attributes.put(SOLD, soldFieldStringMode ? SOLD_BOOKED_TEXT : TRUE);

		JsonArray updates = getUpdatesJson(queryInstance.getObjectID(), attributes);

		try {

			updatePlotRequest(updates, loggingPrefix);

		} catch (IOException e) {

			log.error(loggingPrefix + "Error setting plot as sold for " + queryInstance + ")\n" + e);
		}
	}

	public void abortPlotBooking(Integer objectID, String loggingPrefix) {

		if (objectID == null) {

			log.warn(loggingPrefix + "Could not abort plot booking due to missing Object ID");
			return;
		}

		log.info(loggingPrefix + "Aborting plot booking");

		Map<String, String> attributes = new HashMap<>();

		attributes.put(LOCK, null);

		JsonArray updates = getUpdatesJson(objectID, attributes);

		try {

			updatePlotRequest(updates, loggingPrefix);

		} catch (IOException e) {

			log.error(loggingPrefix + "Error aborting plot booking\n" + e);
		}
	}

	public PlotQueryInstance lockPlot(User user, PlotQueryInstance queryInstance, String loggingPrefix) throws PlotLockException {

		log.info(loggingPrefix + "Locking plot for " + queryInstance);

		Map<String, String> attributes = new HashMap<>();

		queryInstance.setTimestamp(getStringFromDate(new Date()));

		attributes.put(LOCK, queryInstance.getLockKey());

		JsonArray updates = getUpdatesJson(queryInstance.getObjectID(), attributes);

		try {

			updatePlotRequest(updates, loggingPrefix);

			SessionUtils.setAttribute(PlotIntegrationConstants.LOCKED_OBJECT_ID, queryInstance.getObjectID(), user.getSession());

		} catch (IOException e) {

			throw new PlotLockException(e);
		}

		return queryInstance;
	}

	public PlotQueryInstance getPlotDataRequest(PlotQueryInstance queryInstance, String loggingPrefix) throws IOException, PlotUnavailableException {

		log.info(loggingPrefix + "Requesting plot data from integration endpoint for " + queryInstance);

		TagReplacer tagReplacer = new TagReplacer(OBJECT_ID_TAG_SOURCE_FACTORY.getTagSource(queryInstance.getObjectID()));

		String url = tagReplacer.replace(searchPlotURL);

		String response;

		if (tokenHandler != null) {

			response = tokenHandler.sendRequest(token -> {

				String urlWithToken = url + "&token=" + token;

				return sendGetPlotDataRequest(urlWithToken, loggingPrefix);

			}, loggingPrefix);

		} else {

			response = sendGetPlotDataRequest(url, loggingPrefix);
		}

		JSONDeserializer<List<Map<String, Map<String, Object>>>> deserialiser = new JSONDeserializer<>();

		List<Map<String, Map<String, Object>>> features;
		try {

			features = deserialiser.deserialize(response, FEATURES, List.class);

		} catch (Exception e) {

			throw new InvalidResponseException(loggingPrefix + "Error deserializing response: " + e.getMessage());
		}

		Map<String, Object> attributes;
		if (CollectionUtils.isEmpty(features) || (attributes = features.get(0).get(ATTRIBUTES)) == null) {

			throw new InvalidResponseException(loggingPrefix + "Could not parse response from endpoint '" + url + "':" + response);
		}

		attributes = upperCaseAttributeKeys(attributes);

		Integer plotID = NumberUtils.toInt(attributes.get(PLOT_ID));
		String propertyUnitDesignation = StringUtils.toString(attributes.get(PROPERTY_UNIT_DESIGNATION));
		String address = StringUtils.toString(attributes.get(ADDRESS));
		Integer area = NumberUtils.toInt(attributes.get(AREA));
		Integer connectionFee = NumberUtils.toInt(attributes.get(CONNECTION_FEE));
		Integer price = NumberUtils.toInt(attributes.get(PRICE));
		String location = StringUtils.toString(attributes.get(LOCATION));
		String comment = StringUtils.toString(attributes.get(COMMENT));
		String remoteLockKey = StringUtils.toString(attributes.get(LOCK));

		queryInstance.setPlotID(plotID);
		queryInstance.setPropertyUnitDesignation(propertyUnitDesignation);
		queryInstance.setAddress(address);
		queryInstance.setArea(area);
		queryInstance.setConnectionFee(connectionFee);
		queryInstance.setPrice(price);
		queryInstance.setLocation(location);
		queryInstance.setComment(comment);
		queryInstance.setRemoteLockKey(remoteLockKey);

		boolean sold = soldFieldStringMode ? !SOLD_FREE_TEXT.equals(attributes.get(SOLD)) : TRUE.equals(StringUtils.toString(attributes.get(SOLD)));
		boolean bookable = ignoreBookableField || TRUE.equals(StringUtils.toString(attributes.get(BOOKABLE)));
		Date bookableDate = getDateFromString(StringUtils.toString(attributes.get(BOOKABLE_DATE)));

		checkPlotAvailability(queryInstance, sold, bookable, bookableDate, remoteLockKey);

		return queryInstance;
	}

	private void checkPlotAvailability(PlotQueryInstance queryInstance, boolean sold, boolean bookable, Date bookableDate, String remoteLockKey) throws PlotUnavailableException {

		if (sold) {

			throw new PlotSoldException();
		}

		if (!bookable) {

			throw new PlotNotBookableException();
		}

		if (bookableDate == null && queryInstance.getQuery().isRequireBookableDate()) {

			throw new PlotCurrentlyNotBookableException();
		}

		if (bookableDate != null && bookableDate.after(new Date())) {

			throw new PlotCurrentlyNotBookableException();
		}

		if (plotIsLocked(queryInstance.getLockKey(), remoteLockKey)) {

			if (queryInstance.getLockKey() != null) {

				throw new PlotLockExpiredException();
			}

			throw new PlotLockedException();
		}
	}

	private boolean plotIsLocked(String currentlockKey, String remoteLockKey) {

		if (StringUtils.isEmpty(remoteLockKey) || remoteLockKey.equals(currentlockKey)) {

			return false;
		}

		Date lockDate = getDateFromLockKey(remoteLockKey);

		if (lockDate == null) {

			return false;
		}

		Instant lockExpiration = lockDate.toInstant().plus(lockTime, ChronoUnit.MINUTES);

		return now().isBefore(lockExpiration);
	}

	protected Instant now() {

		return Instant.now();
	}

	private void updatePlotRequest(JsonArray updates, String loggingPrefix) throws IOException {

		String data = "f=json&updates=" + updates.toJson();

		if (tokenHandler != null) {

			tokenHandler.sendRequest(token -> {

				String dataWithToken = data + "&token=" + token;

				return sendUpdatePlotRequest(updatePlotURL, dataWithToken, loggingPrefix);

			}, loggingPrefix);

		} else {

			sendUpdatePlotRequest(updatePlotURL, data, loggingPrefix);
		}
	}

	private JsonArray getUpdatesJson(Integer objectID, Map<String, String> extraAttributes) {

		JsonArray updates = new JsonArray();
		JsonObject update = new JsonObject();
		JsonObject attributes = new JsonObject();

		attributes.putField(OBJECT_ID, objectID);

		if (!CollectionUtils.isEmpty(extraAttributes)) {

			for (Map.Entry<String, String> entry : extraAttributes.entrySet()) {

				attributes.putField(entry.getKey(), entry.getValue());
			}
		}

		update.putField(ATTRIBUTES, attributes);

		updates.addNode(update);

		return updates;
	}

	private Map<String, Object> upperCaseAttributeKeys(Map<String, Object> attributes) {

		Map<String, Object> attributesWithUpperCaseKey = new HashMap<>(attributes.size());

		for (Entry<String, Object> entry : attributes.entrySet()) {

			attributesWithUpperCaseKey.put(entry.getKey().toUpperCase(), entry.getValue());
		}

		return attributesWithUpperCaseKey;
	}

	protected String sendUpdatePlotRequest(String url, String data, String loggingPrefix) throws IOException {

		InputStream inputStream = new ByteArrayInputStream(data.getBytes());

		SimpleRequest simpleRequest = new SimpleRequest(url);

		simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
		simpleRequest.setReadTimeout(readTimeout * 1000);

		simpleRequest.addHeader("Content-Type", UPDATE_REQUEST_CONTENT_TYPE);

		if (!StringUtils.isEmpty(refererHeader)) {

			simpleRequest.addHeader("Referer", refererHeader);
		}

		if (verboseAPILoggingEnabled) {

			log.info(loggingPrefix + "Request URL: " + simpleRequest);
			log.info(loggingPrefix + "Request Data: " + data);
		}

		String response = HTTPUtils.sendHTTPPostRequest(simpleRequest, inputStream, INTEGRATION_CHARSET);

		if (verboseAPILoggingEnabled) {

			log.info(loggingPrefix + "Response: " + response);
		}

		return response;
	}

	protected String sendGetPlotDataRequest(String url, String loggingPrefix) throws IOException {

		SimpleRequest simpleRequest = new SimpleRequest(url);

		simpleRequest.setConnectionTimeout(connectionTimeout * 1000);
		simpleRequest.setReadTimeout(readTimeout * 1000);

		if (!StringUtils.isEmpty(refererHeader)) {

			simpleRequest.addHeader("Referer", refererHeader);
		}

		if (verboseAPILoggingEnabled) {

			log.info(loggingPrefix + "Request URL: " + simpleRequest);
		}

		String response = HTTPUtils.sendHTTPGetRequest(simpleRequest, INTEGRATION_CHARSET).getValue();

		if (verboseAPILoggingEnabled) {

			log.info(loggingPrefix + "Response: " + response);
		}

		return response;
	}

	private Date getDateFromString(String date) {

		if (StringUtils.isEmpty(date)) {

			return null;
		}

		Long unixTimestamp = NumberUtils.toLong(date);

		if (unixTimestamp != null) {

			return new Date(unixTimestamp);
		}

		try {

			return DateUtils.DATE_TIME_SECONDS_FORMATTER.parse(date);

		} catch (ParseException e) {

			try {

				return DateUtils.DATE_FORMATTER.parse(date);

			} catch (ParseException e2) {

				return null;
			}
		}
	}

	private String getStringFromDate(Date date) {

		if (date == null) {

			return null;
		}

		return DateUtils.DATE_TIME_SECONDS_FORMATTER.format(date);
	}

	private Date getDateFromLockKey(String lockKey) {

		if (StringUtils.isEmpty(lockKey)) {

			return null;
		}

		return getDateFromString(lockKey.split("\\|")[0]);
	}

}
