package com.nordicpeak.flowengine.queries.plotquery.eventmodule;

import java.io.IOException;
import java.sql.SQLException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeParseException;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import se.unlogic.hierarchy.core.annotations.EventListener;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.beans.SimpleForegroundModuleResponse;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.enums.EventSource;
import se.unlogic.hierarchy.core.exceptions.URINotFoundException;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.utils.HierarchyAnnotatedDAOFactory;
import se.unlogic.hierarchy.foregroundmodules.rest.AnnotatedRESTModule;
import se.unlogic.hierarchy.foregroundmodules.rest.RESTMethod;
import se.unlogic.hierarchy.foregroundmodules.staticcontent.StaticContentModule;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonArray;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.json.JsonObject;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.xml.XMLUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.RequestUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.queries.plotquery.beans.PlotQueryInstance;
import com.nordicpeak.flowengine.queries.plotquery.daos.PlotEventDAO;
import com.nordicpeak.flowengine.queries.plotquery.eventmodule.beans.PlotEvent;
import com.nordicpeak.flowengine.queries.plotquery.eventmodule.beans.PlotEventType;
import com.nordicpeak.flowengine.queries.plotquery.eventmodule.events.PlotHistoryEvent;

public class PlotEventModule extends AnnotatedRESTModule {

	@InstanceManagerDependency(required = true)
	private StaticContentModule staticContentModule;

	private PlotEventDAO plotEventDAO;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		super.createDAOs(dataSource);

		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, PlotEventModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.warn(upgradeResult.toString());
		}

		HierarchyAnnotatedDAOFactory daoFactory = new HierarchyAnnotatedDAOFactory(dataSource, systemInterface);

		plotEventDAO = new PlotEventDAO(dataSource, PlotEvent.class, daoFactory);
	}

	@Override
	public ForegroundModuleResponse defaultMethod(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws URINotFoundException, SQLException {

		Document doc = createDocument(req, uriParser);
		Element listPlotEventsElement = doc.createElement("ListPlotEvents");
		doc.getFirstChild().appendChild(listPlotEventsElement);

		log.warn("User " + user + " listing all plot events");

		return new SimpleForegroundModuleResponse(doc);
	}

	@RESTMethod(alias = "plotevents", method = { "get" }, requireLogin = true)
	public ForegroundModuleResponse getPlotEvents(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws SQLException, IOException {

		log.info("User " + user + " requesting plot events");

		String fromDateParam = req.getParameter("fromDate");
		String toDateParam = req.getParameter("toDate");

		JsonObject jsonObject = new JsonObject(1);

		LocalDateTime fromDate = null;
		LocalDateTime toDate = null;
		try {

			if (!StringUtils.isEmpty(fromDateParam)) {

				fromDate = LocalDate.parse(fromDateParam).atStartOfDay();
			}

			if (!StringUtils.isEmpty(toDateParam)) {

				toDate = LocalDate.parse(toDateParam).atTime(LocalTime.MAX);
			}

		} catch (DateTimeParseException e) {

			res.setStatus(HttpServletResponse.SC_BAD_REQUEST);

			HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
			return null;
		}

		List<PlotEvent> plotEvents = plotEventDAO.getInDateRange(fromDate, toDate);

		if (!CollectionUtils.isEmpty(plotEvents)) {

			JsonArray plotEventsArray = new JsonArray(plotEvents.size());

			for (PlotEvent plotEvent : plotEvents) {

				JsonNode json = plotEvent.toJSON(systemInterface.getUserHandler());

				plotEventsArray.addNode(json);
			}

			jsonObject.putField("data", plotEventsArray);

		} else {

			jsonObject.putField("data", new JsonArray());
		}

		HTTPUtils.sendReponse(jsonObject.toJson(), JsonUtils.getContentType(), res);
		return null;
	}

	@EventListener(channel = PlotQueryInstance.class)
	public void processEvent(PlotHistoryEvent event, EventSource eventSource) throws SQLException {

		PlotQueryInstance queryInstance = event.getQueryInstance();
		User user = event.getUser();
		PlotEventType eventType = event.getEventType();
		String details = event.getDetails();

		createPlotEvent(queryInstance, user, eventType, details);
	}

	public void createPlotEvent(PlotQueryInstance queryInstance, User user, PlotEventType eventType, String details) throws SQLException {

		Integer objectID = queryInstance.getObjectID();
		String propertyUnitDesignation = queryInstance.getPropertyUnitDesignation();
		Integer userID = user.getUserID();

		PlotEvent plotEvent = new PlotEvent(userID, objectID, propertyUnitDesignation, eventType, details);

		log.info("Creating new plotEvent " + plotEvent);

		plotEventDAO.add(plotEvent);
	}

	public Document createDocument(HttpServletRequest req, URIParser uriParser) {

		Document doc = XMLUtils.createDomDocument();
		Element document = doc.createElement("Document");
		document.appendChild(RequestUtils.getRequestInfoAsXML(doc, req, uriParser));
		document.appendChild(this.moduleDescriptor.toXML(doc));

		XMLUtils.appendNewElement(doc, document, "ModuleURI", req.getContextPath() + getFullAlias());
		XMLUtils.appendNewElement(doc, document, "StaticContentURL", staticContentModule.getModuleContentURL(moduleDescriptor));

		doc.appendChild(document);

		return doc;
	}

}