package com.nordicpeak.flowengine.remoteflowinstance;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.net.SocketTimeoutException;
import java.util.List;

import org.apache.log4j.Level;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.TextFieldSettingDescriptor;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.utils.ModuleUtils;
import se.unlogic.hierarchy.core.utils.UserUtils;
import se.unlogic.hierarchy.foregroundmodules.AnnotatedForegroundModule;
import se.unlogic.standardutils.io.CloseUtils;
import se.unlogic.standardutils.serialization.FilteredObjectInputStream;
import se.unlogic.standardutils.time.MillisecondTimeUnits;
import se.unlogic.standardutils.validation.StringIntegerValidator;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.SimpleRequest;
import se.unlogic.webutils.populators.StringHTTPURLPopulator;

import com.nordicpeak.flowengine.Constants;
import com.nordicpeak.flowengine.UserFlowInstanceModule;
import com.nordicpeak.flowengine.beans.FlowInstance;
import com.nordicpeak.flowengine.flowinstanceadmin2.FlowInstanceAdmin2Module;
import com.nordicpeak.flowengine.flowinstanceadmin2.beans.UserManagableFlowInstanceCounts;
import com.nordicpeak.flowengine.flowinstanceadmin2.enums.InstanceFilter;
import com.nordicpeak.flowengine.flowinstanceadmin2.interfaces.ManagableFlowInstanceProvider;
import com.nordicpeak.flowengine.interfaces.UserFlowInstanceProvider;

public class HTTPRemoteFlowInstanceClientModule extends AnnotatedForegroundModule implements ManagableFlowInstanceProvider, UserFlowInstanceProvider {

	private FlowInstanceAdmin2Module flowInstanceAdmin2Module;

	private UserFlowInstanceModule userFlowInstanceModule;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Remote flow instance module URL", description = "URL to the remote flow instance module", required = true, formatValidator = StringHTTPURLPopulator.class)
	private String remoteFlowInstanceModuleURL;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Username", description = "The username to use for the connection", required = true)
	private String username;

	@ModuleSetting(allowsNull = true)
	@TextFieldSettingDescriptor(name = "Password", description = "The password to use for the connection", required = true)
	private String password;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Connection Timeout", description = "Connection timeout in seconds", formatValidator = StringIntegerValidator.class, required = true)
	protected Integer connectionTimeout = 5;

	@ModuleSetting
	@TextFieldSettingDescriptor(name = "Read Timeout", description = "Read timeout in seconds", formatValidator = StringIntegerValidator.class, required = true)
	protected Integer readTimeout = 20;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Register in flow instance admin module ", description = "Controls if this module should register itself in the flow instance admin module")
	private boolean registerAsAdminFlowInstanceProvider;

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Register in user flow instance module", description = "Controls if this module should register itself in the user flow instance module")
	private boolean registerAsUserFlowInstanceProvider;

	private boolean properlyConfigured;

	@Override
	protected void moduleConfigured() throws Exception {

		properlyConfigured = ModuleUtils.checkRequiredModuleSettings(moduleDescriptor, this, systemInterface, Level.ERROR);

		super.moduleConfigured();
	}

	@Override
	public void unload() throws Exception {

		setFlowInstanceAdmin2Module(null);
		setUserFlowInstanceModule(null);

		super.unload();
	}

	@Override
	public UserManagableFlowInstanceCounts getFlowInstanceCounts(User user) {
		
		if (!properlyConfigured) {
			return null;
		}
		
		String username = user.getUsername();

		if (username != null) {

			ByteArrayOutputStream outputStream = null;
			FilteredObjectInputStream inputStream = null;

			try {
				outputStream = new ByteArrayOutputStream();

				SimpleRequest request = new SimpleRequest(remoteFlowInstanceModuleURL + "/managerflowinstances/" + username);

				request.setUsername(this.username);
				request.setPassword(password);
				request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
				request.setReadTimeout(readTimeout * MillisecondTimeUnits.SECOND);

				HTTPUtils.sendHTTPGetRequest(request, outputStream);

				inputStream = new FilteredObjectInputStream(new ByteArrayInputStream(outputStream.toByteArray()), UserManagableFlowInstanceCounts.class);

				UserManagableFlowInstanceCounts flowInstanceCounts = (UserManagableFlowInstanceCounts) inputStream.readObject();

				return flowInstanceCounts;

			} catch (SocketTimeoutException e) {
				
				log.warn("Error getting manager flow instance counts for user " + user + " from " + remoteFlowInstanceModuleURL, e);
			
			} catch (Exception e) {

				log.error("Error getting manager flow instance counts for user " + user + " from " + remoteFlowInstanceModuleURL, e);

			} finally {

				CloseUtils.close(outputStream);
				CloseUtils.close(inputStream);
			}
		}

		return null;
	}
	
	@Override
	public List<FlowInstance> getFlowInstances(InstanceFilter instanceFilter, User user) {

		if (!properlyConfigured) {
			return null;
		}

		String username = user.getUsername();

		if (username != null) {

			ByteArrayOutputStream outputStream = null;
			FilteredObjectInputStream inputStream = null;

			try {
				outputStream = new ByteArrayOutputStream();

				SimpleRequest request = new SimpleRequest(remoteFlowInstanceModuleURL + "/managerflowinstances/" + username + "/" + instanceFilter.name());

				request.setUsername(this.username);
				request.setPassword(password);
				request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
				request.setReadTimeout(readTimeout * MillisecondTimeUnits.SECOND);

				HTTPUtils.sendHTTPGetRequest(request, outputStream);

				inputStream = new FilteredObjectInputStream(new ByteArrayInputStream(outputStream.toByteArray()), HTTPRemoteFlowInstanceShared.ALLOWED_CLASSES);

				@SuppressWarnings("unchecked")
				List<FlowInstance> flowInstances = (List<FlowInstance>) inputStream.readObject();

				return flowInstances;

			} catch (SocketTimeoutException e) {

				log.warn("Error getting manager flow instances for user " + user + " from " + remoteFlowInstanceModuleURL, e);

			} catch (Exception e) {

				log.error("Error getting manager flow instances for user " + user + " from " + remoteFlowInstanceModuleURL, e);

			} finally {

				CloseUtils.close(outputStream);
				CloseUtils.close(inputStream);
			}
		}

		return null;
	}

	@Override
	public List<FlowInstance> getUserFlowInstances(User user) {

		if (!properlyConfigured) {

			return null;
		}

		String citizenIdentifier = UserUtils.getAttribute(Constants.USER_CITIZEN_IDENTIFIER_ATTRIBUTE, user);

		if (citizenIdentifier != null) {

			ByteArrayOutputStream outputStream = null;
			FilteredObjectInputStream inputStream = null;

			try {
				outputStream = new ByteArrayOutputStream();

				SimpleRequest request = new SimpleRequest(remoteFlowInstanceModuleURL + "/userflowinstances/" + citizenIdentifier);

				request.setUsername(username);
				request.setPassword(password);
				request.setConnectionTimeout(connectionTimeout * MillisecondTimeUnits.SECOND);
				request.setReadTimeout(readTimeout * MillisecondTimeUnits.SECOND);

				HTTPUtils.sendHTTPGetRequest(request, outputStream);

				inputStream = new FilteredObjectInputStream(new ByteArrayInputStream(outputStream.toByteArray()), HTTPRemoteFlowInstanceShared.ALLOWED_CLASSES);

				@SuppressWarnings("unchecked")
				List<FlowInstance> flowInstances = (List<FlowInstance>) inputStream.readObject();

				if (flowInstances != null) {

					for (FlowInstance flowInstance : flowInstances) {

						flowInstance.setRemote(true);
					}
				}

				return flowInstances;

			} catch (SocketTimeoutException e) {

				log.warn("Error getting user flow instances for user " + user + " from " + remoteFlowInstanceModuleURL, e);

			}catch (Exception e) {

				log.error("Error getting user flow instances for user " + user + " from " + remoteFlowInstanceModuleURL, e);

			} finally {

				CloseUtils.close(outputStream);
				CloseUtils.close(inputStream);
			}
		}

		return null;
	}

	@InstanceManagerDependency
	public void setFlowInstanceAdmin2Module(FlowInstanceAdmin2Module flowInstanceAdmin2Module) {

		if (flowInstanceAdmin2Module != null && registerAsAdminFlowInstanceProvider) {

			flowInstanceAdmin2Module.addFlowInstanceProvider(this);

		} else if (this.flowInstanceAdmin2Module != null) {

			this.flowInstanceAdmin2Module.removeFlowInstanceProvider(this);
		}

		this.flowInstanceAdmin2Module = flowInstanceAdmin2Module;
	}

	@InstanceManagerDependency
	public void setUserFlowInstanceModule(UserFlowInstanceModule userFlowInstanceModule) {

		if (userFlowInstanceModule != null && registerAsUserFlowInstanceProvider) {

			userFlowInstanceModule.addFlowInstanceProvider(this);

		} else if (this.userFlowInstanceModule != null) {

			this.userFlowInstanceModule.removeFlowInstanceProvider(this);
		}

		this.userFlowInstanceModule = userFlowInstanceModule;
	}
}
