package com.nordicpeak.flowengine.evaluators.tableattributesummaryevaluator;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.numbers.NumberStringyfier;
import se.unlogic.standardutils.numbers.NumberUtils;
import se.unlogic.standardutils.xml.XMLGenerator;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.evaluators.baseevaluator.BaseEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryHandler;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.flowengine.queries.basetablequery.BaseTableColumn;
import com.nordicpeak.flowengine.queries.basetablequery.BaseTableQuery;
import com.nordicpeak.flowengine.queries.dynamictablequery.DropDownAlternative;
import com.nordicpeak.flowengine.queries.dynamictablequery.DynamicTableColumn;
import com.nordicpeak.flowengine.queries.dynamictablequery.DynamicTableQuery;
import com.nordicpeak.flowengine.queries.dynamictablequery.DynamicTableQueryInstance;
import com.nordicpeak.flowengine.queries.dynamictablequery.DynamicTableRow;

public class TableAttributeSummaryEvaluationProviderModule extends BaseEvaluationProviderModule<TableAttributeSummaryEvaluator> {

	protected TableAttributeSummaryEvaluatorCRUD evaluatorCRUD;

	protected AnnotatedDAO<TableAttributeSummaryEvaluator> evaluatorDAO;

	protected QueryParameterFactory<TableAttributeSummaryEvaluator, Integer> evaluatorIDParamFactory;

	@InstanceManagerDependency(required = true)
	protected QueryHandler queryHandler;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, TableAttributeSummaryEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(TableAttributeSummaryEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new TableAttributeSummaryEvaluatorCRUD(TableAttributeSummaryEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, TableAttributeSummaryEvaluator evaluator, EvaluationCallback callback, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {

		if (!(queryInstance instanceof DynamicTableQueryInstance)) {

			log.warn("Query instance " + queryInstance + " is not of the type " + DynamicTableQueryInstance.class.getName());

			return null;
		}
		
		DynamicTableQueryInstance dynamicTableQueryInstance = (DynamicTableQueryInstance) queryInstance;
		List<DynamicTableColumn> columns = dynamicTableQueryInstance.getQuery().getColumns();
		
		if (!CollectionUtils.isEmpty(columns)) {
			
			boolean typeColumnFound = false;
			boolean valueColumnFound = false;
			
			for (DynamicTableColumn column : columns) {
				
				if (column.getColumnID().equals(evaluator.getTypeColumnID())) {
					
					typeColumnFound = true;
				}
				
				if (column.getColumnID().equals(evaluator.getValueColumnID())) {
					
					valueColumnFound = true;
				}
				
				if (typeColumnFound && valueColumnFound) {
					
					break;
				}
			}
			
			if (!typeColumnFound) {
				
				log.warn("Evaluator " + evaluator + " either has no type column specified, or the specified type column has been deleted, aborting evaluation");
				
				return null;
			}
			
			if (!valueColumnFound) {
				
				log.warn("Evaluator " + evaluator + " either has no value column specified, or the specified value column has been deleted, aborting evaluation");
				
				return null;
			}

			
			List<DynamicTableRow> rows = dynamicTableQueryInstance.getRows();
			Set<String> setAttributes = null;
			
			if (!CollectionUtils.isEmpty(rows)) {
				
				Map<Integer, BigDecimal> alternativeValues = new HashMap<Integer, BigDecimal>();
				
				for (DynamicTableRow row : rows) {
					
					Integer alternativeID = NumberUtils.toInt(dynamicTableQueryInstance.getValue(row.getRowIdentifier(), evaluator.getTypeColumnID()));
					BigDecimal value = NumberUtils.toBigDecimalLocaleIndependent(dynamicTableQueryInstance.getValue(row.getRowIdentifier(), evaluator.getValueColumnID()));
					
					if (alternativeID != null && value != null) {
						
						BigDecimal prevValue = alternativeValues.get(alternativeID);
						
						if (prevValue != null) {
							
							alternativeValues.put(alternativeID, prevValue.add(value));
							
						} else {
							
							alternativeValues.put(alternativeID, value);
						}
					}
				}
				
				for (Map.Entry<Integer, BigDecimal> entry : alternativeValues.entrySet()) {
					
					String attributeName = getAttribute(evaluator, entry.getKey());
					BigDecimal attributeValue = entry.getValue();

					if (attributeName != null && attributeValue != null) {
						
						Integer decimals = dynamicTableQueryInstance.getQuery().getDecimalsToShow();
						
						if (decimals != null) {
			
							attributeHandler.setAttribute(attributeName, new NumberStringyfier(decimals, decimals, false).format(attributeValue));
							
						} else {
							
							if (attributeValue.scale() > 0) {
								
								attributeHandler.setAttribute(attributeName, BaseTableColumn.NUMBER_STRINGIFYER_WITH_DECIMALS.format(attributeValue));
								
							} else {
								
								attributeHandler.setAttribute(attributeName, BaseTableColumn.NUMBER_STRINGIFYER_WITHOUT_DECIMALS.format(attributeValue));
							}
						}
						
						setAttributes = CollectionUtils.addAndInstantiateIfNeeded(setAttributes, attributeName);
					}
				}
				
			}
			
			if (!CollectionUtils.isEmpty(evaluator.getAlternativeAttributes())) {
				
				for (AlternativeAttribute alternativeAttribute : evaluator.getAlternativeAttributes()) {
					
					if (setAttributes == null || !setAttributes.contains(alternativeAttribute.getAttribute())) {
						
						attributeHandler.removeAttribute(alternativeAttribute.getAttribute());
					}
				}
			}
		}

		return null;
	}

	private String getAttribute(TableAttributeSummaryEvaluator evaluator, Integer alternativeID) {

		if (alternativeID != null && evaluator != null) {

			List<AlternativeAttribute> alternativeAttributes = evaluator.getAlternativeAttributes();

			if (!CollectionUtils.isEmpty(alternativeAttributes)) {

				for (AlternativeAttribute alternativeAttribute : alternativeAttributes) {

					if (alternativeAttribute.getAlternativeID().equals(alternativeID)) {

						return alternativeAttribute.getAttribute();
					}
				}
			}
		}

		return null;
	}

	@Override
	public Evaluator createEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		TableAttributeSummaryEvaluator evaluator = new TableAttributeSummaryEvaluator();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());

		return evaluator;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return DynamicTableQuery.class.isAssignableFrom(queryClass);
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		TableAttributeSummaryEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<TableAttributeSummaryEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());

		List<BaseTableColumn> sourceColumns = ((BaseTableQuery) sourceQuery).getColumns();
		List<BaseTableColumn> copyColumns = ((BaseTableQuery) copyQuery).getColumns();

		if (!CollectionUtils.isEmpty(sourceColumns) && !CollectionUtils.isEmpty(copyColumns)) {

			for (int i = 0; i < sourceColumns.size(); i++) {

				if (sourceColumns.get(i).getColumnID().equals(evaluator.getTypeColumnID())) {

					evaluator.setTypeColumnID(copyColumns.get(i).getColumnID());

					List<DropDownAlternative> sourceAlternatives = sourceColumns.get(i).getDropDownAlternatives();
					List<DropDownAlternative> copyAlternatives = copyColumns.get(i).getDropDownAlternatives();

					setAlternativeAttributeIDs(evaluator, sourceAlternatives, copyAlternatives);
				}

				if (sourceColumns.get(i).getColumnID().equals(evaluator.getValueColumnID())) {

					evaluator.setValueColumnID(copyColumns.get(i).getColumnID());
				}
			}
		}

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	private void setAlternativeAttributeIDs(TableAttributeSummaryEvaluator evaluator, List<DropDownAlternative> sourceAlternatives, List<DropDownAlternative> copyAlternatives) {

		if (!CollectionUtils.isEmpty(sourceAlternatives) && !CollectionUtils.isEmpty(copyAlternatives)) {

			for (int i = 0; i < sourceAlternatives.size(); i++) {

				for (AlternativeAttribute alternativeAttribute : evaluator.getAlternativeAttributes()) {

					if (sourceAlternatives.get(i).getAlternativeID().equals(alternativeAttribute.getAlternativeID())) {

						alternativeAttribute.setAlternativeID(copyAlternatives.get(i).getAlternativeID());
					}
				}
			}
		}
	}

	protected Class<TableAttributeSummaryEvaluator> getEvaluatorClass() {

		return TableAttributeSummaryEvaluator.class;
	}

	protected AnnotatedDAO<TableAttributeSummaryEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	protected QueryParameterFactory<TableAttributeSummaryEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return evaluatorIDParamFactory;
	}

	@Override
	public Evaluator getEvaluator(MutableEvaluatorDescriptor descriptor, InstanceMetadata instanceMetadata) throws Throwable {

		TableAttributeSummaryEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<TableAttributeSummaryEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())));

		if (evaluator != null) {

			evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());
		}

		return evaluator;
	}

	@Override
	public boolean deleteEvaluator(ImmutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		evaluatorDAO.delete(new HighLevelQuery<TableAttributeSummaryEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())), transactionHandler);

		return true;
	}

	public QueryHandler getQueryHandler() {

		return this.queryHandler;
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public TableAttributeSummaryEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		TableAttributeSummaryEvaluator evaluator = getEvaluatorClass().newInstance();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		evaluator.populate(descriptor.getImportParser().getNode(XMLGenerator.getElementName(evaluator.getClass())));

		if (query instanceof DynamicTableQuery && ((BaseTableQuery) query).getColumnConversionMap() != null) {

			Map<Integer, Integer> columnConversionMap = ((BaseTableQuery) query).getColumnConversionMap();
			Map<Integer, Integer> dropDownAlternativeConversionMap = ((BaseTableQuery) query).getDropDownAlternativeConversionMap();

			if (evaluator.getTypeColumnID() != null) {

				Integer newTypeColumnID = columnConversionMap.get(evaluator.getTypeColumnID());

				evaluator.setTypeColumnID(newTypeColumnID);

				if (!CollectionUtils.isEmpty(evaluator.getAlternativeAttributes())) {

					for (AlternativeAttribute alternativeAttribute : evaluator.getAlternativeAttributes()) {

						Integer newAlternativeID = dropDownAlternativeConversionMap.get(alternativeAttribute.getAlternativeID());

						alternativeAttribute.setAlternativeID(newAlternativeID);
					}
				}
			}

			if (evaluator.getValueColumnID() != null) {

				Integer newValueColumnID = columnConversionMap.get(evaluator.getValueColumnID());

				evaluator.setValueColumnID(newValueColumnID);
			}

		}

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}
}
