package com.nordicpeak.flowengine.evaluators.useraddressquerystateevaluator;

import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.CheckboxSettingDescriptor;
import se.unlogic.hierarchy.core.annotations.InstanceManagerDependency;
import se.unlogic.hierarchy.core.annotations.ModuleSetting;
import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.AttributeHandler;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.collections.CollectionUtils;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.json.JsonNode;
import se.unlogic.standardutils.json.JsonUtils;
import se.unlogic.standardutils.object.ObjectUtils;
import se.unlogic.webutils.http.HTTPUtils;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.evaluators.basequerystateevaluator.BaseQueryStateEvaluationProviderModule;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;
import com.nordicpeak.lmsearch.LMSearchProvider;

public class UserAddressQueryStateEvaluationProviderModule extends BaseQueryStateEvaluationProviderModule<UserAddressQueryStateEvaluator> {

	@ModuleSetting
	@CheckboxSettingDescriptor(name = "Enable selection by street number", description = "Enables selecting addresses with specific street numbers")
	protected boolean selectByStreetNumber;
	
	protected AnnotatedDAO<UserAddressQueryStateEvaluator> evaluatorDAO;

	protected QueryParameterFactory<UserAddressQueryStateEvaluator, Integer> evaluatorIDParamFactory;

	protected UserAddressQueryStateEvaluatorCRUD evaluatorCRUD;

	@InstanceManagerDependency(required = true)
	private LMSearchProvider lmSearchProvider;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, UserAddressQueryStateEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(UserAddressQueryStateEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new UserAddressQueryStateEvaluatorCRUD(UserAddressQueryStateEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, UserAddressQueryStateEvaluator evaluator, EvaluationCallback callback, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {
		
		if (evaluator.isInitialized()) {
			
			return null;
		}

		if (poster != null) {
			
			evaluator.setInitialized(true);

			AttributeHandler posterAttributeHandler = poster.getAttributeHandler();

			String posterAddress = posterAttributeHandler.getString("official.address");
			String posterPostalAddress = posterAttributeHandler.getString("official.postalAddress");

			if (ObjectUtils.containsNull(posterAddress, posterPostalAddress)) {

				log.info("User " + poster + " has no offical address");
				return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
			}

			boolean addressMatches = false;

			if (!CollectionUtils.isEmpty(evaluator.getAddresses())) {
			
				for (Address address : evaluator.getAddresses()) {
	
					if (posterPostalAddress.equalsIgnoreCase(address.getPostalAddress()) && posterAddress.toLowerCase().startsWith(address.getStreet().toLowerCase())) {
	
						addressMatches = true;
						break;
					}
				}
			
			}

			if ((addressMatches || evaluator.isInverted()) && !(addressMatches && evaluator.isInverted())) {

				return applyEvaluatorQueryStates(queryInstance, evaluator, callback, attributeHandler);

			} else {

				return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
			}

		} else {

			return restoreDefaultQueryStates(queryInstance, evaluator, callback, attributeHandler);
		}

	}

	@Override
	public Evaluator createEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		UserAddressQueryStateEvaluator evaluator = new UserAddressQueryStateEvaluator();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());

		return evaluator;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return true;
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		UserAddressQueryStateEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<UserAddressQueryStateEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());
		
		if (!CollectionUtils.isEmpty(evaluator.getAddresses())) {
			
			for (Address address : evaluator.getAddresses()) {
				
				address.setAddressID(null);
			}
		}

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	@Override
	protected Class<UserAddressQueryStateEvaluator> getEvaluatorClass() {

		return UserAddressQueryStateEvaluator.class;
	}

	@Override
	protected AnnotatedDAO<UserAddressQueryStateEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	@Override
	protected QueryParameterFactory<UserAddressQueryStateEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return this.evaluatorIDParamFactory;
	}

	@Override
	protected List<Integer> getTargetQueryIDs(UserAddressQueryStateEvaluator evaluator, EvaluationCallback callback) {

		return Collections.singletonList(callback.getQueryInstance().getQueryInstanceDescriptor().getQueryDescriptor().getQueryID());
	}

	@Override
	public UserAddressQueryStateEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		UserAddressQueryStateEvaluator evaluator = importEvaluatorDescriptor(descriptor, transactionHandler, query, statusConversionMap);

		evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}

	@WebPublic(toLowerCase = true)
	public ForegroundModuleResponse searchForAddress(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Throwable {

		if (lmSearchProvider == null) {

			log.warn("User " + user + " tried to search for an address but search provider is not configured");
			return null;
		}

		JsonNode data = lmSearchProvider.searchForAddress(req, res, user, uriParser);

		HTTPUtils.sendReponse(data.toJson(), JsonUtils.getContentType(), res);

		return null;
	}

	public boolean isSelectByStreetNumber() {

		return selectByStreetNumber;
	}

}
