package com.nordicpeak.flowengine.evaluators.setweightbyexistingflowinstancestatusevaluator;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

import se.unlogic.hierarchy.core.annotations.WebPublic;
import se.unlogic.hierarchy.core.beans.User;
import se.unlogic.hierarchy.core.interfaces.ForegroundModuleResponse;
import se.unlogic.hierarchy.core.interfaces.attributes.MutableAttributeHandler;
import se.unlogic.standardutils.dao.AnnotatedDAO;
import se.unlogic.standardutils.dao.HighLevelQuery;
import se.unlogic.standardutils.dao.QueryParameterFactory;
import se.unlogic.standardutils.dao.SimpleAnnotatedDAOFactory;
import se.unlogic.standardutils.dao.TransactionHandler;
import se.unlogic.standardutils.dao.querys.BooleanQuery;
import se.unlogic.standardutils.db.tableversionhandler.TableVersionHandler;
import se.unlogic.standardutils.db.tableversionhandler.UpgradeResult;
import se.unlogic.standardutils.db.tableversionhandler.XMLDBScriptProvider;
import se.unlogic.standardutils.string.StringUtils;
import se.unlogic.standardutils.xml.XMLGenerator;
import se.unlogic.webutils.http.URIParser;

import com.nordicpeak.flowengine.beans.EvaluationResponse;
import com.nordicpeak.flowengine.beans.QueryModification;
import com.nordicpeak.flowengine.enums.ContentType;
import com.nordicpeak.flowengine.evaluators.baseevaluator.BaseEvaluationProviderModule;
import com.nordicpeak.flowengine.evaluators.setweightevaluator.SetWeightEvaluationEvent;
import com.nordicpeak.flowengine.interfaces.EvaluationCallback;
import com.nordicpeak.flowengine.interfaces.Evaluator;
import com.nordicpeak.flowengine.interfaces.ImmutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.ImmutableFlowInstance;
import com.nordicpeak.flowengine.interfaces.ImmutableStatus;
import com.nordicpeak.flowengine.interfaces.InstanceMetadata;
import com.nordicpeak.flowengine.interfaces.MutableEvaluatorDescriptor;
import com.nordicpeak.flowengine.interfaces.Query;
import com.nordicpeak.flowengine.interfaces.QueryInstance;

public class SetWeightByExistingFlowInstanceStatusEvaluationProviderModule extends BaseEvaluationProviderModule<SetWeightByExistingFlowInstanceStatusEvaluator> {

	protected SetWeightByExistingFlowInstanceStatusEvaluatorCRUD evaluatorCRUD;

	protected AnnotatedDAO<SetWeightByExistingFlowInstanceStatusEvaluator> evaluatorDAO;

	protected QueryParameterFactory<SetWeightByExistingFlowInstanceStatusEvaluator, Integer> evaluatorIDParamFactory;

	@Override
	protected void createDAOs(DataSource dataSource) throws Exception {

		//Automatic table version handling
		UpgradeResult upgradeResult = TableVersionHandler.upgradeDBTables(dataSource, SetWeightByExistingFlowInstanceStatusEvaluationProviderModule.class.getName(), new XMLDBScriptProvider(this.getClass().getResourceAsStream("DB script.xml")));

		if (upgradeResult.isUpgrade()) {

			log.info(upgradeResult.toString());
		}

		evaluatorDAO = new SimpleAnnotatedDAOFactory(dataSource).getDAO(SetWeightByExistingFlowInstanceStatusEvaluator.class);
		evaluatorIDParamFactory = evaluatorDAO.getParamFactory("evaluatorID", Integer.class);

		evaluatorCRUD = new SetWeightByExistingFlowInstanceStatusEvaluatorCRUD(SetWeightByExistingFlowInstanceStatusEvaluator.class, evaluatorDAO.getWrapper("evaluatorID", Integer.class), this);
	}

	@WebPublic(alias = "config")
	public ForegroundModuleResponse configureEvaluator(HttpServletRequest req, HttpServletResponse res, User user, URIParser uriParser) throws Exception {

		return this.evaluatorCRUD.update(req, res, user, uriParser);
	}

	@Override
	public EvaluationResponse evaluate(QueryInstance queryInstance, User user, User poster, SetWeightByExistingFlowInstanceStatusEvaluator evaluator, EvaluationCallback callback, boolean hasQueryValidationErrors, MutableAttributeHandler attributeHandler) {

		if (user == null || evaluator.getContentTypes() == null || callback.getFlowInstance().getFirstSubmitted() != null) {

			return null;
		}

		BigDecimal weight = BigDecimal.ZERO;

		try {

			log.info("Checking for existing flow instances in family of flow " + callback.getFlowInstance().getFlow() + " for poster " + user + " with status content types " + StringUtils.toCommaSeparatedString(evaluator.getContentTypes()));

			if (hasDuplicates(user, callback.getFlowInstance(), evaluator.getContentTypes())) {

				weight = BigDecimal.valueOf(evaluator.getWeight());
			}

		} catch (Exception e) {

			log.error("Error checking for existing flow instances " + callback.getFlowInstance(), e);
		}

		List<QueryModification> queryModifications = callback.sendEvent(new SetWeightEvaluationEvent(evaluator.getEvaluatorID(), evaluator.getWeightType(), weight, attributeHandler), true);

		if (queryModifications != null) {

			return new EvaluationResponse(queryModifications);
		}

		return null;
	}

	private boolean hasDuplicates(User user, ImmutableFlowInstance flowInstance, List<ContentType> contentTypes) throws SQLException {

		BooleanQuery query;

		StringBuilder sql = new StringBuilder();

		sql.append("SELECT flowengine_flow_instances.flowInstanceID ");
		sql.append("FROM flowengine_flow_instances ");
		sql.append("INNER JOIN flowengine_flows ON flowengine_flows.flowID = flowengine_flow_instances.flowID ");
		sql.append("INNER JOIN flowengine_flow_statuses ON flowengine_flow_statuses.statusID = flowengine_flow_instances.statusID ");
		sql.append("WHERE ");
		sql.append("flowengine_flow_instances.poster = ? ");
		sql.append("AND flowengine_flows.flowFamilyID = ? ");

		sql.append("AND flowengine_flow_statuses.contentType IN (");
		sql.append(contentTypes.stream().map(c -> "?").collect(Collectors.joining(", ")));
		sql.append(") ");

		if (flowInstance.getFlowInstanceID() != null) {

			sql.append("AND flowengine_flow_instances.flowInstanceID != ? ");
		}

		sql.append("LIMIT 1");

		query = new BooleanQuery(dataSource, sql.toString());
		int paramCount = 1;

		query.setInt(paramCount++, user.getUserID());
		query.setInt(paramCount++, flowInstance.getFlow().getFlowFamily().getFlowFamilyID());

		for (ContentType contentType : contentTypes) {

			query.setString(paramCount++, contentType.toString());
		}

		if (flowInstance.getFlowInstanceID() != null) {

			query.setInt(paramCount++, flowInstance.getFlowInstanceID());
		}

		return query.executeQuery();
	}

	@Override
	public Evaluator createEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		SetWeightByExistingFlowInstanceStatusEvaluator evaluator = new SetWeightByExistingFlowInstanceStatusEvaluator();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());

		return evaluator;
	}

	@Override
	public boolean supportsQueryType(Class<? extends Query> queryClass) {

		return true;
	}

	@Override
	public void copyEvaluator(MutableEvaluatorDescriptor sourceEvaluatorDescriptor, MutableEvaluatorDescriptor copyEvaluatorDescriptor, Query sourceQuery, Query copyQuery, Map<Integer, ImmutableStatus> statusConversionMap, TransactionHandler transactionHandler) throws SQLException {

		SetWeightByExistingFlowInstanceStatusEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<SetWeightByExistingFlowInstanceStatusEvaluator>(evaluatorIDParamFactory.getParameter(sourceEvaluatorDescriptor.getEvaluatorID())));

		evaluator.setEvaluatorID(copyEvaluatorDescriptor.getEvaluatorID());

		this.evaluatorDAO.add(evaluator, transactionHandler, null);
	}

	protected Class<SetWeightByExistingFlowInstanceStatusEvaluator> getEvaluatorClass() {

		return SetWeightByExistingFlowInstanceStatusEvaluator.class;
	}

	protected AnnotatedDAO<SetWeightByExistingFlowInstanceStatusEvaluator> getEvaluatorDAO() {

		return evaluatorDAO;
	}

	protected QueryParameterFactory<SetWeightByExistingFlowInstanceStatusEvaluator, Integer> getEvaluatorIDParameterFactory() {

		return evaluatorIDParamFactory;
	}

	@Override
	public Evaluator getEvaluator(MutableEvaluatorDescriptor descriptor, InstanceMetadata instanceMetadata) throws Throwable {

		SetWeightByExistingFlowInstanceStatusEvaluator evaluator = evaluatorDAO.get(new HighLevelQuery<SetWeightByExistingFlowInstanceStatusEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())));

		if (evaluator != null) {

			evaluator.init(descriptor, getFullAlias() + "/config/" + descriptor.getEvaluatorID());
		}

		return evaluator;
	}

	@Override
	public boolean deleteEvaluator(ImmutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler) throws Throwable {

		evaluatorDAO.delete(new HighLevelQuery<SetWeightByExistingFlowInstanceStatusEvaluator>(evaluatorIDParamFactory.getParameter(descriptor.getEvaluatorID())), transactionHandler);

		return true;
	}

	@Override
	public SetWeightByExistingFlowInstanceStatusEvaluator importEvaluator(MutableEvaluatorDescriptor descriptor, TransactionHandler transactionHandler, Query query, Map<Integer, ImmutableStatus> statusConversionMap) throws Throwable {

		SetWeightByExistingFlowInstanceStatusEvaluator evaluator = getEvaluatorClass().newInstance();

		evaluator.setEvaluatorID(descriptor.getEvaluatorID());

		evaluator.populate(descriptor.getImportParser().getNode(XMLGenerator.getElementName(evaluator.getClass())));

		this.evaluatorDAO.add(evaluator, transactionHandler, null);

		return evaluator;
	}

}
